# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Parse and convert amc motion capture data."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control.mujoco import math as mjmath
import numpy as np
from scipy import interpolate
from six.moves import range

MOCAP_DT = 1.0/120.0
CONVERSION_LENGTH = 0.056444

_CMU_MOCAP_JOINT_ORDER = (
    'root0', 'root1', 'root2', 'root3', 'root4', 'root5', 'lowerbackrx',
    'lowerbackry', 'lowerbackrz', 'upperbackrx', 'upperbackry', 'upperbackrz',
    'thoraxrx', 'thoraxry', 'thoraxrz', 'lowerneckrx', 'lowerneckry',
    'lowerneckrz', 'upperneckrx', 'upperneckry', 'upperneckrz', 'headrx',
    'headry', 'headrz', 'rclaviclery', 'rclaviclerz', 'rhumerusrx',
    'rhumerusry', 'rhumerusrz', 'rradiusrx', 'rwristry', 'rhandrx', 'rhandrz',
    'rfingersrx', 'rthumbrx', 'rthumbrz', 'lclaviclery', 'lclaviclerz',
    'lhumerusrx', 'lhumerusry', 'lhumerusrz', 'lradiusrx', 'lwristry',
    'lhandrx', 'lhandrz', 'lfingersrx', 'lthumbrx', 'lthumbrz', 'rfemurrx',
    'rfemurry', 'rfemurrz', 'rtibiarx', 'rfootrx', 'rfootrz', 'rtoesrx',
    'lfemurrx', 'lfemurry', 'lfemurrz', 'ltibiarx', 'lfootrx', 'lfootrz',
    'ltoesrx'
)

Converted = collections.namedtuple('Converted',
                                   ['qpos', 'qvel', 'time'])


def convert(file_name, physics, timestep):
  """Converts the parsed .amc values into qpos and qvel values and resamples.

  Args:
    file_name: The .amc file to be parsed and converted.
    physics: The corresponding physics instance.
    timestep: Desired output interval between resampled frames.

  Returns:
    A namedtuple with fields:
        `qpos`, a numpy array containing converted positional variables.
        `qvel`, a numpy array containing converted velocity variables.
        `time`, a numpy array containing the corresponding times.
  """
  frame_values = parse(file_name)
  joint2index = {}
  for name in physics.named.data.qpos.axes.row.names:
    joint2index[name] = physics.named.data.qpos.axes.row.convert_key_item(name)
  index2joint = {}
  for joint, index in joint2index.items():
    if isinstance(index, slice):
      indices = range(index.start, index.stop)
    else:
      indices = [index]
    for ii in indices:
      index2joint[ii] = joint

  # Convert frame_values to qpos
  amcvals2qpos_transformer = Amcvals2qpos(index2joint, _CMU_MOCAP_JOINT_ORDER)
  qpos_values = []
  for frame_value in frame_values:
    qpos_values.append(amcvals2qpos_transformer(frame_value))
  qpos_values = np.stack(qpos_values)  # Time by nq

  # Interpolate/resample.
  # Note: interpolate quaternions rather than euler angles (slerp).
  # see https://en.wikipedia.org/wiki/Slerp
  qpos_values_resampled = []
  time_vals = np.arange(0, len(frame_values)*MOCAP_DT - 1e-8, MOCAP_DT)
  time_vals_new = np.arange(0, len(frame_values)*MOCAP_DT, timestep)
  while time_vals_new[-1] > time_vals[-1]:
    time_vals_new = time_vals_new[:-1]

  for i in range(qpos_values.shape[1]):
    f = interpolate.splrep(time_vals, qpos_values[:, i])
    qpos_values_resampled.append(interpolate.splev(time_vals_new, f))

  qpos_values_resampled = np.stack(qpos_values_resampled)  # nq by ntime

  qvel_list = []
  for t in range(qpos_values_resampled.shape[1]-1):
    p_tp1 = qpos_values_resampled[:, t + 1]
    p_t = qpos_values_resampled[:, t]
    qvel = [(p_tp1[:3]-p_t[:3])/ timestep,
            mjmath.mj_quat2vel(
                mjmath.mj_quatdiff(p_t[3:7], p_tp1[3:7]), timestep),
            (p_tp1[7:]-p_t[7:])/ timestep]
    qvel_list.append(np.concatenate(qvel))

  qvel_values_resampled = np.vstack(qvel_list).T

  return Converted(qpos_values_resampled, qvel_values_resampled, time_vals_new)


def parse(file_name):
  """Parses the amc file format."""
  values = []
  fid = open(file_name, 'r')
  line = fid.readline().strip()
  frame_ind = 1
  first_frame = True
  while True:
    # Parse first frame.
    if first_frame and line[0] == str(frame_ind):
      first_frame = False
      frame_ind += 1
      frame_vals = []
      while True:
        line = fid.readline().strip()
        if not line or line == str(frame_ind):
          values.append(np.array(frame_vals, dtype=np.float))
          break
        tokens = line.split()
        frame_vals.extend(tokens[1:])
    # Parse other frames.
    elif line == str(frame_ind):
      frame_ind += 1
      frame_vals = []
      while True:
        line = fid.readline().strip()
        if not line or line == str(frame_ind):
          values.append(np.array(frame_vals, dtype=np.float))
          break
        tokens = line.split()
        frame_vals.extend(tokens[1:])
    else:
      line = fid.readline().strip()
      if not line:
        break
  return values


class Amcvals2qpos(object):
  """Callable that converts .amc values for a frame and to MuJoCo qpos format.
  """

  def __init__(self, index2joint, joint_order):
    """Initializes a new Amcvals2qpos instance.

    Args:
      index2joint: List of joint angles in .amc file.
      joint_order: List of joint names in MuJoco MJCF.
    """
    # Root is x,y,z, then quat.
    # need to get indices of qpos that order for amc default order
    self.qpos_root_xyz_ind = [0, 1, 2]
    self.root_xyz_ransform = np.array(
        [[1, 0, 0], [0, 0, -1], [0, 1, 0]]) * CONVERSION_LENGTH
    self.qpos_root_quat_ind = [3, 4, 5, 6]
    amc2qpos_transform = np.zeros((len(index2joint), len(joint_order)))
    for i in range(len(index2joint)):
      for j in range(len(joint_order)):
        if index2joint[i] == joint_order[j]:
          if 'rx' in index2joint[i]:
            amc2qpos_transform[i][j] = 1
          elif 'ry' in index2joint[i]:
            amc2qpos_transform[i][j] = 1
          elif 'rz' in index2joint[i]:
            amc2qpos_transform[i][j] = 1
    self.amc2qpos_transform = amc2qpos_transform

  def __call__(self, amc_val):
    """Converts a `.amc` frame to MuJoCo qpos format."""
    amc_val_rad = np.deg2rad(amc_val)
    qpos = np.dot(self.amc2qpos_transform, amc_val_rad)

    # Root.
    qpos[:3] = np.dot(self.root_xyz_ransform, amc_val[:3])
    qpos_quat = mjmath.euler2quat(amc_val[3], amc_val[4], amc_val[5])
    qpos_quat = mjmath.mj_quatprod(mjmath.euler2quat(90, 0, 0), qpos_quat)

    for i, ind in enumerate(self.qpos_root_quat_ind):
      qpos[ind] = qpos_quat[i]

    return qpos



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for the action scale wrapper."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.
from absl.testing import absltest
from absl.testing import parameterized
from dm_control.rl import control
from dm_control.suite.wrappers import action_scale
from dm_env import specs
import mock
import numpy as np


def make_action_spec(lower=(-1.,), upper=(1.,)):
  lower, upper = np.broadcast_arrays(lower, upper)
  return specs.BoundedArray(
      shape=lower.shape, dtype=float, minimum=lower, maximum=upper)


def make_mock_env(action_spec):
  action_spec = action_spec
  env = mock.Mock(spec=control.Environment)
  env.action_spec.return_value = action_spec
  return env


class ActionScaleTest(parameterized.TestCase):

  def assertStepCalledOnceWithCorrectAction(self, env, expected_action):
    # NB: `assert_called_once_with()` doesn't support numpy arrays.
    env.step.assert_called_once()
    actual_action = env.step.call_args_list[0][0][0]
    np.testing.assert_array_equal(expected_action, actual_action)

  @parameterized.parameters(
      {
          'minimum': np.r_[-1., -1.],
          'maximum': np.r_[1., 1.],
          'scaled_minimum': np.r_[-2., -2.],
          'scaled_maximum': np.r_[2., 2.],
      },
      {
          'minimum': np.r_[-2., -2.],
          'maximum': np.r_[2., 2.],
          'scaled_minimum': np.r_[-1., -1.],
          'scaled_maximum': np.r_[1., 1.],
      },
      {
          'minimum': np.r_[-1., -1.],
          'maximum': np.r_[1., 1.],
          'scaled_minimum': np.r_[-2., -2.],
          'scaled_maximum': np.r_[1., 1.],
      },
      {
          'minimum': np.r_[-1., -1.],
          'maximum': np.r_[1., 1.],
          'scaled_minimum': np.r_[-1., -1.],
          'scaled_maximum': np.r_[2., 2.],
      },
  )
  def test_step(self, minimum, maximum, scaled_minimum, scaled_maximum):
    action_spec = make_action_spec(lower=minimum, upper=maximum)
    env = make_mock_env(action_spec=action_spec)
    wrapped_env = action_scale.Wrapper(
        env, minimum=scaled_minimum, maximum=scaled_maximum)

    time_step = wrapped_env.step(scaled_minimum)
    self.assertStepCalledOnceWithCorrectAction(env, minimum)
    self.assertIs(time_step, env.step(minimum))

    env.reset_mock()

    time_step = wrapped_env.step(scaled_maximum)
    self.assertStepCalledOnceWithCorrectAction(env, maximum)
    self.assertIs(time_step, env.step(maximum))

  @parameterized.parameters(
      {
          'minimum': np.r_[-1., -1.],
          'maximum': np.r_[1., 1.],
      },
      {
          'minimum': np.r_[0, 1],
          'maximum': np.r_[2, 3],
      },
  )
  def test_correct_action_spec(self, minimum, maximum):
    original_action_spec = make_action_spec(
        lower=np.r_[-2., -2.], upper=np.r_[2., 2.])
    env = make_mock_env(action_spec=original_action_spec)
    wrapped_env = action_scale.Wrapper(env, minimum=minimum, maximum=maximum)
    new_action_spec = wrapped_env.action_spec()
    np.testing.assert_array_equal(new_action_spec.minimum, minimum)
    np.testing.assert_array_equal(new_action_spec.maximum, maximum)

  @parameterized.parameters('reset', 'observation_spec', 'control_timestep')
  def test_method_delegated_to_underlying_env(self, method_name):
    env = make_mock_env(action_spec=make_action_spec())
    wrapped_env = action_scale.Wrapper(env, minimum=0, maximum=1)
    env_method = getattr(env, method_name)
    wrapper_method = getattr(wrapped_env, method_name)
    out = wrapper_method()
    env_method.assert_called_once_with()
    self.assertIs(out, env_method())

  def test_invalid_action_spec_type(self):
    action_spec = [make_action_spec()] * 2
    env = make_mock_env(action_spec=action_spec)
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        action_scale._ACTION_SPEC_MUST_BE_BOUNDED_ARRAY.format(action_spec)):
      action_scale.Wrapper(env, minimum=0, maximum=1)

  @parameterized.parameters(
      {'name': 'minimum', 'bounds': np.r_[np.nan]},
      {'name': 'minimum', 'bounds': np.r_[-np.inf]},
      {'name': 'maximum', 'bounds': np.r_[np.inf]},
  )
  def test_non_finite_bounds(self, name, bounds):
    kwargs = {'minimum': np.r_[-1.], 'maximum': np.r_[1.]}
    kwargs[name] = bounds
    env = make_mock_env(action_spec=make_action_spec())
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        action_scale._MUST_BE_FINITE.format(name=name, bounds=bounds)):
      action_scale.Wrapper(env, **kwargs)

  @parameterized.parameters(
      {'name': 'minimum', 'bounds': np.r_[1., 2., 3.]},
      {'name': 'minimum', 'bounds': np.r_[[1.], [2.], [3.]]},
  )
  def test_invalid_bounds_shape(self, name, bounds):
    shape = (2,)
    kwargs = {'minimum': np.zeros(shape), 'maximum': np.ones(shape)}
    kwargs[name] = bounds
    action_spec = make_action_spec(lower=[-1, -1], upper=[2, 3])
    env = make_mock_env(action_spec=action_spec)
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        action_scale._MUST_BROADCAST.format(
            name=name, bounds=bounds, shape=shape)):
      action_scale.Wrapper(env, **kwargs)

if __name__ == '__main__':
  absltest.main()



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Wrapper control suite environments that adds Gaussian noise to actions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import dm_env
import numpy as np


_BOUNDS_MUST_BE_FINITE = (
    'All bounds in `env.action_spec()` must be finite, got: {action_spec}')


class Wrapper(dm_env.Environment):
  """Wraps a control environment and adds Gaussian noise to actions."""

  def __init__(self, env, scale=0.01):
    """Initializes a new action noise Wrapper.

    Args:
      env: The control suite environment to wrap.
      scale: The standard deviation of the noise, expressed as a fraction
        of the max-min range for each action dimension.

    Raises:
      ValueError: If any of the action dimensions of the wrapped environment are
        unbounded.
    """
    action_spec = env.action_spec()
    if not (np.all(np.isfinite(action_spec.minimum)) and
            np.all(np.isfinite(action_spec.maximum))):
      raise ValueError(_BOUNDS_MUST_BE_FINITE.format(action_spec=action_spec))
    self._minimum = action_spec.minimum
    self._maximum = action_spec.maximum
    self._noise_std = scale * (action_spec.maximum - action_spec.minimum)
    self._env = env

  def step(self, action):
    noisy_action = action + self._env.task.random.normal(scale=self._noise_std)
    # Clip the noisy actions in place so that they fall within the bounds
    # specified by the `action_spec`. Note that MuJoCo implicitly clips out-of-
    # bounds control inputs, but we also clip here in case the actions do not
    # correspond directly to MuJoCo actuators, or if there are other wrapper
    # layers that expect the actions to be within bounds.
    np.clip(noisy_action, self._minimum, self._maximum, out=noisy_action)
    return self._env.step(noisy_action)

  def reset(self):
    return self._env.reset()

  def observation_spec(self):
    return self._env.observation_spec()

  def action_spec(self):
    return self._env.action_spec()

  def __getattr__(self, name):
    return getattr(self._env, name)



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Utilities for testing environment hooks."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import contextlib
import inspect

from dm_control import composer
from dm_control import mjcf
from six.moves import range


def add_bodies_and_actuators(mjcf_model, num_actuators):
  if num_actuators % 2:
    raise ValueError('num_actuators is not a multiple of 2')
  for _ in range(num_actuators // 2):
    body = mjcf_model.worldbody.add('body')
    body.add('inertial', pos=[0, 0, 0], mass=1, diaginertia=[1, 1, 1])
    joint_x = body.add('joint', axis=[1, 0, 0])
    mjcf_model.actuator.add('position', joint=joint_x)
    joint_y = body.add('joint', axis=[0, 1, 0])
    mjcf_model.actuator.add('position', joint=joint_y)


class HooksTracker(object):
  """Helper class for tracking call order of callbacks."""

  def __init__(self, test_case, physics_timestep, control_timestep,
               *args, **kwargs):
    super(HooksTracker, self).__init__(*args, **kwargs)
    self.tracked = False
    self._test_case = test_case
    self._call_count = collections.defaultdict(lambda: 0)
    self._physics_timestep = physics_timestep
    self._physics_steps_per_control_step = (
        round(int(control_timestep / physics_timestep)))

    mro = inspect.getmro(type(self))
    self._has_super = mro[mro.index(HooksTracker) + 1] != object

  def assertEqual(self, actual, expected, msg=''):
    msg = '{}: {}: {!r} != {!r}'.format(type(self), msg, actual, expected)
    self._test_case.assertEqual(actual, expected, msg)

  def assertHooksNotCalled(self, *hook_names):
    for hook_name in hook_names:
      self.assertEqual(
          self._call_count[hook_name], 0,
          'assertHooksNotCalled: hook_name = {!r}'.format(hook_name))

  def assertHooksCalledOnce(self, *hook_names):
    for hook_name in hook_names:
      self.assertEqual(
          self._call_count[hook_name], 1,
          'assertHooksCalledOnce: hook_name = {!r}'.format(hook_name))

  def assertCompleteEpisode(self, control_steps):
    self.assertHooksCalledOnce('initialize_episode_mjcf',
                               'after_compile',
                               'initialize_episode')
    physics_steps = control_steps * self._physics_steps_per_control_step
    self.assertEqual(self._call_count['before_step'], control_steps)
    self.assertEqual(self._call_count['before_substep'], physics_steps)
    self.assertEqual(self._call_count['after_substep'], physics_steps)
    self.assertEqual(self._call_count['after_step'], control_steps)

  def assertPhysicsStepCountEqual(self, physics, expected_count):
    actual_count = int(round(physics.time() / self._physics_timestep))
    self.assertEqual(actual_count, expected_count)

  def reset_call_counts(self):
    self._call_count = collections.defaultdict(lambda: 0)

  def initialize_episode_mjcf(self, random_state):
    """Implements `initialize_episode_mjcf` Composer callback."""
    if self._has_super:
      super(HooksTracker, self).initialize_episode_mjcf(random_state)
    if not self.tracked:
      return
    self.assertHooksNotCalled('after_compile',
                              'initialize_episode',
                              'before_step',
                              'before_substep',
                              'after_substep',
                              'after_step')
    self._call_count['initialize_episode_mjcf'] += 1

  def after_compile(self, physics, random_state):
    """Implements `after_compile` Composer callback."""
    if self._has_super:
      super(HooksTracker, self).after_compile(physics, random_state)
    if not self.tracked:
      return
    self.assertHooksCalledOnce('initialize_episode_mjcf')
    self.assertHooksNotCalled('initialize_episode',
                              'before_step',
                              'before_substep',
                              'after_substep',
                              'after_step')
    # Number of physics steps is always consistent with `before_substep`.
    self.assertPhysicsStepCountEqual(physics,
                                     self._call_count['before_substep'])
    self._call_count['after_compile'] += 1

  def initialize_episode(self, physics, random_state):
    """Implements `initialize_episode` Composer callback."""
    if self._has_super:
      super(HooksTracker, self).initialize_episode(physics, random_state)
    if not self.tracked:
      return
    self.assertHooksCalledOnce('initialize_episode_mjcf',
                               'after_compile')
    self.assertHooksNotCalled('before_step',
                              'before_substep',
                              'after_substep',
                              'after_step')
    # Number of physics steps is always consistent with `before_substep`.
    self.assertPhysicsStepCountEqual(physics,
                                     self._call_count['before_substep'])
    self._call_count['initialize_episode'] += 1

  def before_step(self, physics, *args):
    """Implements `before_step` Composer callback."""
    if self._has_super:
      super(HooksTracker, self).before_step(physics, *args)
    if not self.tracked:
      return
    self.assertHooksCalledOnce('initialize_episode_mjcf',
                               'after_compile',
                               'initialize_episode')

    # `before_step` is only called in between complete control steps.
    self.assertEqual(
        self._call_count['after_step'], self._call_count['before_step'])

    # Complete control steps imply complete physics steps.
    self.assertEqual(
        self._call_count['after_substep'], self._call_count['before_substep'])

    # Number of physics steps is always consistent with `before_substep`.
    self.assertPhysicsStepCountEqual(physics,
                                     self._call_count['before_substep'])

    self._call_count['before_step'] += 1

  def before_substep(self, physics, *args):
    """Implements `before_substep` Composer callback."""
    if self._has_super:
      super(HooksTracker, self).before_substep(physics, *args)
    if not self.tracked:
      return
    self.assertHooksCalledOnce('initialize_episode_mjcf',
                               'after_compile',
                               'initialize_episode')

    # We are inside a partial control step, so `after_step` should lag behind.
    self.assertEqual(
        self._call_count['after_step'], self._call_count['before_step'] - 1)

    # `before_substep` is only called in between complete physics steps.
    self.assertEqual(
        self._call_count['after_substep'], self._call_count['before_substep'])

    # Number of physics steps is always consistent with `before_substep`.
    self.assertPhysicsStepCountEqual(
        physics, self._call_count['before_substep'])

    self._call_count['before_substep'] += 1

  def after_substep(self, physics, random_state):
    """Implements `after_substep` Composer callback."""
    if self._has_super:
      super(HooksTracker, self).after_substep(physics, random_state)
    if not self.tracked:
      return
    self.assertHooksCalledOnce('initialize_episode_mjcf',
                               'after_compile',
                               'initialize_episode')

    # We are inside a partial control step, so `after_step` should lag behind.
    self.assertEqual(
        self._call_count['after_step'], self._call_count['before_step'] - 1)

    # We are inside a partial physics step, so `after_substep` should be behind.
    self.assertEqual(self._call_count['after_substep'],
                     self._call_count['before_substep'] - 1)

    # Number of physics steps is always consistent with `before_substep`.
    self.assertPhysicsStepCountEqual(
        physics, self._call_count['before_substep'])

    self._call_count['after_substep'] += 1

  def after_step(self, physics, random_state):
    """Implements `after_step` Composer callback."""
    if self._has_super:
      super(HooksTracker, self).after_step(physics, random_state)
    if not self.tracked:
      return
    self.assertHooksCalledOnce('initialize_episode_mjcf',
                               'after_compile',
                               'initialize_episode')

    # We are inside a partial control step, so `after_step` should lag behind.
    self.assertEqual(
        self._call_count['after_step'], self._call_count['before_step'] - 1)

    # `after_step` is only called in between complete physics steps.
    self.assertEqual(
        self._call_count['after_substep'], self._call_count['before_substep'])

    # Number of physics steps is always consistent with `before_substep`.
    self.assertPhysicsStepCountEqual(
        physics, self._call_count['before_substep'])

    # Check that the number of physics steps is consistent with control steps.
    self.assertEqual(
        self._call_count['before_substep'],
        self._call_count['before_step'] * self._physics_steps_per_control_step)

    self._call_count['after_step'] += 1


class TrackedEntity(HooksTracker, composer.Entity):
  """A `composer.Entity` that tracks call order of callbacks."""

  def _build(self, name):
    self._mjcf_root = mjcf.RootElement(model=name)

  @property
  def mjcf_model(self):
    return self._mjcf_root

  @property
  def name(self):
    return self._mjcf_root.model


class TrackedTask(HooksTracker, composer.NullTask):
  """A `composer.Task` that tracks call order of callbacks."""

  def __init__(self, physics_timestep, control_timestep, *args, **kwargs):
    super(TrackedTask, self).__init__(physics_timestep=physics_timestep,
                                      control_timestep=control_timestep,
                                      *args, **kwargs)
    self.set_timesteps(physics_timestep=physics_timestep,
                       control_timestep=control_timestep)
    add_bodies_and_actuators(self.root_entity.mjcf_model, num_actuators=4)


class HooksTestMixin(object):
  """A mixin for an `absltest.TestCase` to track call order of callbacks."""

  def setUp(self):
    """Sets up the test case."""
    super(HooksTestMixin, self).setUp()

    self.num_episodes = 5
    self.steps_per_episode = 100

    self.control_timestep = 0.05
    self.physics_timestep = 0.002

    self.extra_hooks = HooksTracker(physics_timestep=self.physics_timestep,
                                    control_timestep=self.control_timestep,
                                    test_case=self)

    self.entities = []
    for i in range(9):
      self.entities.append(TrackedEntity(name='entity_{}'.format(i),
                                         physics_timestep=self.physics_timestep,
                                         control_timestep=self.control_timestep,
                                         test_case=self))

    ########################################
    # Make the following entity hierarchy  #
    #                  0                   #
    #       1          2          3        #
    #     4   5      6   7                 #
    #     8                                #
    ########################################

    self.entities[4].attach(self.entities[8])
    self.entities[1].attach(self.entities[4])
    self.entities[1].attach(self.entities[5])
    self.entities[0].attach(self.entities[1])

    self.entities[2].attach(self.entities[6])
    self.entities[2].attach(self.entities[7])
    self.entities[0].attach(self.entities[2])

    self.entities[0].attach(self.entities[3])

    self.task = TrackedTask(root_entity=self.entities[0],
                            physics_timestep=self.physics_timestep,
                            control_timestep=self.control_timestep,
                            test_case=self)

  @contextlib.contextmanager
  def track_episode(self):
    tracked_objects = [self.task, self.extra_hooks] + self.entities
    for obj in tracked_objects:
      obj.reset_call_counts()
      obj.tracked = True
    yield
    for obj in tracked_objects:
      obj.assertCompleteEpisode(self.steps_per_episode)
      obj.tracked = False



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Mujoco Physics viewer, with custom input controllers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control.mujoco.wrapper import mjbindings
from dm_control.viewer import renderer
from dm_control.viewer import user_input
from dm_control.viewer import util
from six.moves import range

constants = mjbindings.constants
enums = mjbindings.enums
functions = mjbindings.functions

_NUM_GROUP_KEYS = 10

_PAN_CAMERA_VERTICAL_MOUSE = user_input.Exclusive(
    user_input.MOUSE_BUTTON_RIGHT)
_PAN_CAMERA_HORIZONTAL_MOUSE = user_input.Exclusive(
    (user_input.MOUSE_BUTTON_RIGHT, user_input.MOD_SHIFT))
_ROTATE_OBJECT_MOUSE = user_input.Exclusive(
    (user_input.MOUSE_BUTTON_LEFT, user_input.MOD_CONTROL))
_MOVE_OBJECT_VERTICAL_MOUSE = user_input.Exclusive(
    (user_input.MOUSE_BUTTON_RIGHT, user_input.MOD_CONTROL))
_MOVE_OBJECT_HORIZONTAL_MOUSE = user_input.Exclusive(
    (user_input.MOUSE_BUTTON_RIGHT, user_input.MOD_SHIFT_CONTROL))

_PAN_CAMERA_VERTICAL_TOUCHPAD = user_input.Exclusive(
    (user_input.MOUSE_BUTTON_LEFT, user_input.MOD_ALT))
_PAN_CAMERA_HORIZONTAL_TOUCHPAD = user_input.Exclusive(
    (user_input.MOUSE_BUTTON_RIGHT, user_input.MOD_ALT))
_ROTATE_OBJECT_TOUCHPAD = user_input.Exclusive(
    user_input.MOUSE_BUTTON_RIGHT)
_MOVE_OBJECT_VERTICAL_TOUCHPAD = user_input.Exclusive(
    (user_input.MOUSE_BUTTON_LEFT, user_input.MOD_CONTROL))
_MOVE_OBJECT_HORIZONTAL_TOUCHPAD = user_input.Exclusive(
    (user_input.MOUSE_BUTTON_LEFT, user_input.MOD_SHIFT_CONTROL))

_ROTATE_CAMERA = user_input.Exclusive(user_input.MOUSE_BUTTON_LEFT)
_CENTER_CAMERA = user_input.DoubleClick(user_input.MOUSE_BUTTON_RIGHT)
_SELECT_OBJECT = user_input.DoubleClick(user_input.MOUSE_BUTTON_LEFT)
_TRACK_OBJECT = user_input.DoubleClick(
    (user_input.MOUSE_BUTTON_RIGHT, user_input.MOD_CONTROL))
_FREE_LOOK = user_input.KEY_ESCAPE
_NEXT_CAMERA = user_input.KEY_RIGHT_BRACKET
_PREVIOUS_CAMERA = user_input.KEY_LEFT_BRACKET
_ZOOM_TO_SCENE = (user_input.KEY_A, user_input.MOD_CONTROL)
_DOUBLE_BUFFERING = user_input.KEY_F5
_PREV_RENDERING_MODE = (user_input.KEY_F6, user_input.MOD_SHIFT)
_NEXT_RENDERING_MODE = user_input.KEY_F6
_PREV_LABELING_MODE = (user_input.KEY_F7, user_input.MOD_SHIFT)
_NEXT_LABELING_MODE = user_input.KEY_F7
_PRINT_CAMERA = user_input.KEY_F11
_VISUALIZATION_FLAGS = user_input.Range(
    [ord(functions.mjVISSTRING[i][2]) for i in
     range(0, enums.mjtVisFlag.mjNVISFLAG)])
_GEOM_GROUPS = user_input.Range(
    [i + ord('0') for i in range(min(_NUM_GROUP_KEYS, constants.mjNGROUP))])
_SITE_GROUPS = user_input.Range(
    [(i + ord('0'), user_input.MOD_SHIFT) for i in
     range(min(_NUM_GROUP_KEYS, constants.mjNGROUP))])
_RENDERING_FLAGS = user_input.Range(
    [ord(functions.mjRNDSTRING[i][2]) for i in
     range(0, enums.mjtRndFlag.mjNRNDFLAG)])

_CAMERA_MOVEMENT_ACTIONS = [enums.mjtMouse.mjMOUSE_MOVE_V,
                            enums.mjtMouse.mjMOUSE_ROTATE_H]

# Translates mouse wheel rotations to zoom speed.
_SCROLL_SPEED_FACTOR = 0.05

# Distance, in meters, at which to focus on the clicked object.
_LOOK_AT_DISTANCE = 1.5


class Viewer(object):
  """Viewport displaying the contents of a physics world."""

  def __init__(self, viewport, mouse, keyboard):
    """Instance initializer.

    Args:
      viewport: Render viewport, instance of renderer.Viewport.
      mouse: A mouse device.
      keyboard: A keyboard device.
    """
    self._viewport = viewport
    self._mouse = mouse

    self._null_perturbation = renderer.NullPerturbation()
    self._render_settings = renderer.RenderSettings()
    self._input_map = user_input.InputMap(mouse, keyboard)

    self._camera = None
    self._camera_settings = None
    self._renderer = None
    self._manipulator = None
    self._free_camera = None
    self._camera_select = None

  def __del__(self):
    del self._camera
    del self._renderer
    del self._manipulator
    del self._free_camera
    del self._camera_select

  def initialize(self, physics, renderer_instance, touchpad):
    """Initialize the viewer.

    Args:
      physics: Physics instance.
      renderer_instance: A renderer.Base instance.
      touchpad: A boolean, use input dedicated to touchpad.
    """
    self._camera = renderer.SceneCamera(
        physics.model, physics.data, self._render_settings,
        self._camera_settings)

    self._manipulator = ManipulationController(
        self._viewport, self._camera, self._mouse)

    self._free_camera = FreeCameraController(
        self._viewport, self._camera, self._mouse, self._manipulator)

    self._camera_select = CameraSelector(
        physics.model, self._camera, self._free_camera)

    self._renderer = renderer_instance

    self._input_map.clear_bindings()

    if touchpad:
      self._input_map.bind(
          self._manipulator.set_move_vertical_mode,
          _MOVE_OBJECT_VERTICAL_TOUCHPAD)
      self._input_map.bind(
          self._manipulator.set_move_horizontal_mode,
          _MOVE_OBJECT_HORIZONTAL_TOUCHPAD)
      self._input_map.bind(
          self._manipulator.set_rotate_mode, _ROTATE_OBJECT_TOUCHPAD)
      self._input_map.bind(
          self._free_camera.set_pan_vertical_mode,
          _PAN_CAMERA_VERTICAL_TOUCHPAD)
      self._input_map.bind(
          self._free_camera.set_pan_horizontal_mode,
          _PAN_CAMERA_HORIZONTAL_TOUCHPAD)
    else:
      self._input_map.bind(
          self._manipulator.set_move_vertical_mode, _MOVE_OBJECT_VERTICAL_MOUSE)
      self._input_map.bind(
          self._manipulator.set_move_horizontal_mode,
          _MOVE_OBJECT_HORIZONTAL_MOUSE)
      self._input_map.bind(
          self._manipulator.set_rotate_mode, _ROTATE_OBJECT_MOUSE)
      self._input_map.bind(
          self._free_camera.set_pan_vertical_mode, _PAN_CAMERA_VERTICAL_MOUSE)
      self._input_map.bind(
          self._free_camera.set_pan_horizontal_mode,
          _PAN_CAMERA_HORIZONTAL_MOUSE)

    self._input_map.bind(self._print_camera_transform, _PRINT_CAMERA)
    self._input_map.bind(
        self._render_settings.select_prev_rendering_mode, _PREV_RENDERING_MODE)
    self._input_map.bind(
        self._render_settings.select_next_rendering_mode, _NEXT_RENDERING_MODE)
    self._input_map.bind(
        self._render_settings.select_prev_labeling_mode, _PREV_LABELING_MODE)
    self._input_map.bind(
        self._render_settings.select_next_labeling_mode, _NEXT_LABELING_MODE)
    self._input_map.bind(
        self._render_settings.select_prev_labeling_mode, _PREV_LABELING_MODE)
    self._input_map.bind(
        self._render_settings.toggle_stereo_buffering, _DOUBLE_BUFFERING)
    self._input_map.bind(
        self._render_settings.toggle_visualization_flag, _VISUALIZATION_FLAGS)
    self._input_map.bind(
        self._render_settings.toggle_site_group, _SITE_GROUPS)
    self._input_map.bind(
        self._render_settings.toggle_geom_group, _GEOM_GROUPS)
    self._input_map.bind(
        self._render_settings.toggle_rendering_flag, _RENDERING_FLAGS)

    self._input_map.bind(self._camera.zoom_to_scene, _ZOOM_TO_SCENE)
    self._input_map.bind(self._camera_select.select_next, _NEXT_CAMERA)
    self._input_map.bind(self._camera_select.select_previous, _PREVIOUS_CAMERA)
    self._input_map.bind_z_axis(self._free_camera.zoom)
    self._input_map.bind_plane(self._free_camera.on_move)
    self._input_map.bind(self._free_camera.set_rotate_mode, _ROTATE_CAMERA)
    self._input_map.bind(self._free_camera.center, _CENTER_CAMERA)
    self._input_map.bind(self._free_camera.track, _TRACK_OBJECT)
    self._input_map.bind(self._free_camera.free_look, _FREE_LOOK)
    self._input_map.bind(self._manipulator.select, _SELECT_OBJECT)
    self._input_map.bind_plane(self._manipulator.on_move)

  def deinitialize(self):
    """Deinitializes the viewer instance."""
    self._input_map.clear_bindings()
    self._camera_settings = self._camera.settings if self._camera else None
    del self._camera
    del self._renderer
    del self._manipulator
    del self._free_camera
    del self._camera_select
    self._camera = None
    self._renderer = None
    self._manipulator = None
    self._free_camera = None
    self._camera_select = None

  def render(self):
    """Renders the visualized scene."""
    if self._camera and self._renderer:  # Can be None during env reload.
      scene = self._camera.render(self.perturbation)
      self._render_settings.apply_settings(scene)
      self._renderer.render(self._viewport, scene)

  def zoom_to_scene(self):
    """Utility method that set the camera to embrace the entire scene."""
    if self._camera:
      self._camera.zoom_to_scene()

  def _print_camera_transform(self):
    if self._camera:
      rotation_mtx, position = self._camera.transform
      right, up, _ = rotation_mtx
      print('<camera pos="%.3f %.3f %.3f" '
            'xyaxes="%.3f %.3f %.3f %.3f %.3f %.3f"/>' % (
                position[0], position[1], position[2], right[0], right[1],
                right[2], up[0], up[1], up[2]))

  @property
  def perturbation(self):
    """Returns an active renderer.Perturbation object."""
    if self._manipulator and self._manipulator.perturbation:
      return self._manipulator.perturbation
    else:
      return self._null_perturbation

  @property
  def camera(self):
    """Returns an active renderer.SceneCamera instance."""
    return self._camera

  @property
  def render_settings(self):
    """Returns renderer.RenderSettings used by this viewer."""
    return self._render_settings


class CameraSelector(object):
  """Binds camera behavior to user input."""

  def __init__(self, model, camera, free_camera, **unused):
    """Instance initializer.

    Args:
      model: Instance of MjModel.
      camera: Instance of SceneCamera.
      free_camera: Instance of FreeCameraController.
      **unused: Other arguments, not used by this class.
    """
    del unused  # Unused.
    self._model = model
    self._camera = camera
    self._free_ctrl = free_camera

    self._camera_idx = -1
    self._active_ctrl = self._free_ctrl

  def select_previous(self):
    """Cycles to the previous scene camera."""
    self._camera_idx -= 1
    if not self._model.ncam or self._camera_idx < -1:
      self._camera_idx = self._model.ncam - 1
    self._commit_selection()

  def select_next(self):
    """Cycles to the next scene camera."""
    self._camera_idx += 1
    if not self._model.ncam or self._camera_idx >= self._model.ncam:
      self._camera_idx = -1
    self._commit_selection()

  def _commit_selection(self):
    """Selects a controller that should go with the selected camera."""
    if self._camera_idx < 0:
      self._activate(self._free_ctrl)
    else:
      self._camera.set_fixed_mode(self._camera_idx)
      self._activate(None)

  def _activate(self, controller):
    """Activates a sub-controller."""
    if controller == self._active_ctrl:
      return

    if self._active_ctrl is not None:
      self._active_ctrl.deactivate()
    self._active_ctrl = controller
    if self._active_ctrl is not None:
      self._active_ctrl.activate()


class FreeCameraController(object):
  """Implements the free camera behavior."""

  def __init__(self, viewport, camera, pointer, selection_service, **unused):
    """Instance initializer.

    Args:
      viewport: Instance of mujoco_viewer.Viewport.
      camera: Instance of mujoco_viewer.SceneCamera.
      pointer: A pointer that moves around the screen and is used to point at
        bodies. Implements a single attribute - 'position' - that returns a
        2-component vector of pointer's screen space position.
      selection_service: An instance of a class implementing a
        'selected_body_id' property.
      **unused: Other optional parameters not used by this class.
    """
    del unused  # Unused.
    self._viewport = viewport
    self._camera = camera
    self._pointer = pointer
    self._selection_service = selection_service
    self._active = True
    self._tracked_body_idx = -1
    self._action = util.AtomicAction()

  def activate(self):
    """Activates the controller."""
    self._active = True
    self._update_camera_mode()

  def deactivate(self):
    """Deactivates the controller."""
    self._active = False
    self._action = util.AtomicAction()

  def set_pan_vertical_mode(self, enable):
    """Starts/ends the camera panning action along the vertical plane.

    Args:
      enable: A boolean flag, True to start the action, False to end it.
    """
    if self._active:
      if enable:
        self._action.begin(enums.mjtMouse.mjMOUSE_MOVE_V)
      else:
        self._action.end(enums.mjtMouse.mjMOUSE_MOVE_V)

  def set_pan_horizontal_mode(self, enable):
    """Starts/ends the camera panning action along the horizontal plane.

    Args:
      enable: A boolean flag, True to start the action, False to end it.
    """
    if self._active:
      if enable:
        self._action.begin(enums.mjtMouse.mjMOUSE_MOVE_H)
      else:
        self._action.end(enums.mjtMouse.mjMOUSE_MOVE_H)

  def set_rotate_mode(self, enable):
    """Starts/ends the camera rotation action.

    Args:
      enable: A boolean flag, True to start the action, False to end it.
    """
    if self._active:
      if enable:
        self._action.begin(enums.mjtMouse.mjMOUSE_ROTATE_H)
      else:
        self._action.end(enums.mjtMouse.mjMOUSE_ROTATE_H)

  def center(self):
    """Focuses camera on the object the pointer is currently pointing at."""
    if self._active:
      body_id, world_pos = self._camera.raycast(self._viewport,
                                                self._pointer.position)
      if body_id >= 0:
        self._camera.look_at(world_pos, _LOOK_AT_DISTANCE)

  def on_move(self, position, translation):
    """Translates mouse moves onto camera movements."""
    del position
    if self._action.in_progress:
      viewport_offset = self._viewport.screen_to_viewport(translation)
      self._camera.move(self._action.watermark, viewport_offset)

  def zoom(self, zoom_factor):
    """Zooms the camera in/out.

    Args:
      zoom_factor: A floating point value, by how much to zoom the camera.
        Positive values zoom the camera in, negative values zoom it out.
    """
    if self._active:
      offset = [0, _SCROLL_SPEED_FACTOR * zoom_factor * -1.]
      self._camera.move(enums.mjtMouse.mjMOUSE_ZOOM, offset)

  def track(self):
    """Makes the camera track the currently selected object.

    The selection is managed by the selection service.
    """
    if self._active and self._tracked_body_idx < 0:
      self._tracked_body_idx = self._selection_service.selected_body_id
      self._update_camera_mode()

  def free_look(self):
    """Switches the camera to a free-look mode."""
    if self._active:
      self._tracked_body_idx = -1
      self._update_camera_mode()

  def _update_camera_mode(self):
    """Sets the camera into a tracking or a free-look mode."""
    if self._tracked_body_idx >= 0:
      self._camera.set_tracking_mode(self._tracked_body_idx)
    else:
      self._camera.set_freelook_mode()


class ManipulationController(object):
  """Binds control over scene objects to user input."""

  def __init__(self, viewport, camera, pointer, **unused):
    """Instance initializer.

    Args:
      viewport: Instance of mujoco_viewer.Viewport.
      camera: Instance of mujoco_viewer.SceneCamera.
      pointer: A pointer that moves around the screen and is used to point at
        bodies. Implements a single attribute - 'position' - that returns a
        2-component vector of pointer's screen space position.
      **unused: Other arguments, unused by this class.
    """
    del unused  # Unused.
    self._viewport = viewport
    self._camera = camera
    self._pointer = pointer
    self._action = util.AtomicAction(self._update_action)
    self._perturb = None

  def select(self):
    """Translates mouse double-clicks to object selection action."""
    body_id, _ = self._camera.raycast(self._viewport, self._pointer.position)
    if body_id >= 0:
      self._perturb = self._camera.new_perturbation(body_id)
    else:
      self._perturb = None

  def set_move_vertical_mode(self, enable):
    """Begins/ends an object translation action along the vertical plane.

    Args:
      enable: A boolean flag, True begins the action, False ends it.
    """
    if enable:
      self._action.begin(enums.mjtMouse.mjMOUSE_MOVE_V)
    else:
      self._action.end(enums.mjtMouse.mjMOUSE_MOVE_V)

  def set_move_horizontal_mode(self, enable):
    """Begins/ends an object translation action along the horizontal plane.

    Args:
      enable: A boolean flag, True begins the action, False ends it.
    """
    if enable:
      self._action.begin(enums.mjtMouse.mjMOUSE_MOVE_H)
    else:
      self._action.end(enums.mjtMouse.mjMOUSE_MOVE_H)

  def set_rotate_mode(self, enable):
    """Begins/ends an object rotation action.

    Args:
      enable: A boolean flag, True begins the action, False ends it.
    """
    if enable:
      self._action.begin(enums.mjtMouse.mjMOUSE_ROTATE_H)
    else:
      self._action.end(enums.mjtMouse.mjMOUSE_ROTATE_H)

  def _update_action(self, action):
    if self._perturb is not None:
      if action is not None:
        _, grab_pos = self._camera.raycast(self._viewport,
                                           self._pointer.position)
        self._perturb.start_move(action, grab_pos)
      else:
        self._perturb.end_move()

  def on_move(self, position, translation):
    """Translates mouse moves to selected object movements."""
    del position
    if self._perturb is not None and self._action.in_progress:
      viewport_offset = self._viewport.screen_to_viewport(translation)
      self._perturb.tick_move(viewport_offset)

  @property
  def perturbation(self):
    """Returns the Perturbation object that represents the manipulated body."""
    return self._perturb

  @property
  def selected_body_id(self):
    """Returns the id of the selected body, or -1 if none is selected."""
    return self._perturb.body_id if self._perturb is not None else -1



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Helper module that specifies the path to Kinova assets."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import six

# pylint: disable=g-import-not-at-top
if six.PY2:
  import imp
  _DM_CONTROL_ROOT = imp.find_module('dm_control')[1]
else:
  import importlib
  _DM_CONTROL_ROOT = os.path.dirname(
      importlib.util.find_spec('dm_control').origin)
# pylint: enable=g-import-not-at-top

KINOVA_ROOT = os.path.join(_DM_CONTROL_ROOT, 'third_party/kinova')



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Randomization functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control.mujoco.wrapper import mjbindings
import numpy as np
from six.moves import range


def random_limited_quaternion(random, limit):
  """Generates a random quaternion limited to the specified rotations."""
  axis = random.randn(3)
  axis /= np.linalg.norm(axis)
  angle = random.rand() * limit

  quaternion = np.zeros(4)
  mjbindings.mjlib.mju_axisAngle2Quat(quaternion, axis, angle)

  return quaternion


def randomize_limited_and_rotational_joints(physics, random=None):
  """Randomizes the positions of joints defined in the physics body.

  The following randomization rules apply:
    - Bounded joints (hinges or sliders) are sampled uniformly in the bounds.
    - Unbounded hinges are samples uniformly in [-pi, pi]
    - Quaternions for unlimited free joints and ball joints are sampled
      uniformly on the unit 3-sphere.
    - Quaternions for limited ball joints are sampled uniformly on a sector
      of the unit 3-sphere.
    - The linear degrees of freedom of free joints are not randomized.

  Args:
    physics: Instance of 'Physics' class that holds a loaded model.
    random: Optional instance of 'np.random.RandomState'. Defaults to the global
      NumPy random state.
  """
  random = random or np.random

  hinge = mjbindings.enums.mjtJoint.mjJNT_HINGE
  slide = mjbindings.enums.mjtJoint.mjJNT_SLIDE
  ball = mjbindings.enums.mjtJoint.mjJNT_BALL
  free = mjbindings.enums.mjtJoint.mjJNT_FREE

  qpos = physics.named.data.qpos

  for joint_id in range(physics.model.njnt):
    joint_name = physics.model.id2name(joint_id, 'joint')
    joint_type = physics.model.jnt_type[joint_id]
    is_limited = physics.model.jnt_limited[joint_id]
    range_min, range_max = physics.model.jnt_range[joint_id]

    if is_limited:
      if joint_type == hinge or joint_type == slide:
        qpos[joint_name] = random.uniform(range_min, range_max)

      elif joint_type == ball:
        qpos[joint_name] = random_limited_quaternion(random, range_max)

    else:
      if joint_type == hinge:
        qpos[joint_name] = random.uniform(-np.pi, np.pi)

      elif joint_type == ball:
        quat = random.randn(4)
        quat /= np.linalg.norm(quat)
        qpos[joint_name] = quat

      elif joint_type == free:
        quat = random.rand(4)
        quat /= np.linalg.norm(quat)
        qpos[joint_name][3:] = quat




# Copyright 2020 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for rescaling bodies."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from dm_control import mjcf
from dm_control.locomotion.walkers import rescale
import numpy as np


class RescaleTest(absltest.TestCase):

  def setUp(self):
    super(RescaleTest, self).setUp()

    # build a simple three-link chain with an endpoint site
    self._mjcf_model = mjcf.RootElement()
    body = self._mjcf_model.worldbody.add('body', pos=[0, 0, 0])
    body.add('geom', type='capsule', fromto=[0, 0, 0, 0, 0, -0.4], size=[0.06])
    body.add('joint', type='ball')
    body = body.add('body', pos=[0, 0, -0.5])
    body.add('geom', type='capsule', pos=[0, 0, -0.15], size=[0.06, 0.15])
    body.add('joint', type='ball')
    body = body.add('body', pos=[0, 0, -0.4])
    body.add('geom', type='capsule', fromto=[0, 0, 0, 0.3, 0, -0.4],
             size=[0.06])
    body.add('joint', type='ball')
    body.add('site', name='endpoint', type='sphere', pos=[0.3, 0, -0.4],
             size=[0.1])

  def test_rescale(self):
    # verify endpoint is where expected
    physics = mjcf.Physics.from_mjcf_model(self._mjcf_model)
    np.testing.assert_allclose(physics.named.data.site_xpos['endpoint'],
                               np.array([0.3, 0., -1.3]), atol=1e-15)

    # rescale chain and verify endpoint is where expected after modification
    subtree_root = self._mjcf_model
    position_factor = .5
    size_factor = .5
    rescale.rescale_subtree(subtree_root, position_factor, size_factor)
    physics = mjcf.Physics.from_mjcf_model(self._mjcf_model)
    np.testing.assert_allclose(physics.named.data.site_xpos['endpoint'],
                               np.array([0.15, 0., -0.65]), atol=1e-15)

if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""A 2x4 Duplo brick."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os

from dm_control import composer
from dm_control import mjcf
from dm_control.composer import define
from dm_control.composer.observation import observable
import numpy as np
from six.moves import range

_DUPLO_XML_PATH = os.path.join(os.path.dirname(__file__), 'duplo2x4.xml')

# Stud radii are drawn from a uniform distribution. The `variation` argument
# scales the minimum and maximum whilst keeping the lower quartile constant.
_StudSize = collections.namedtuple(
    '_StudSize', ['minimum', 'lower_quartile', 'maximum'])
_StudParams = collections.namedtuple('_StudParams', ['easy_align', 'flanges'])

_STUD_SIZE_PARAMS = {
    _StudParams(easy_align=False, flanges=False):
        _StudSize(minimum=0.004685, lower_quartile=0.004781, maximum=0.004898),
    _StudParams(easy_align=False, flanges=True):
        _StudSize(minimum=0.004609, lower_quartile=0.004647, maximum=0.004716),
    _StudParams(easy_align=True, flanges=False):
        _StudSize(minimum=0.004754, lower_quartile=0.004844, maximum=0.004953),
    _StudParams(easy_align=True, flanges=True):
        _StudSize(minimum=0.004695, lower_quartile=0.004717, maximum=0.004765)
}

_COLOR_NOT_BETWEEN_0_AND_1 = (
    'All values in `color` must be between 0 and 1, got {!r}.')


class Duplo(composer.Entity):
  """A 2x4 Duplo brick."""

  def _build(self, easy_align=False, flanges=True, variation=0.0,
             color=(1., 0., 0.)):
    """Initializes a new `Duplo` instance.

    Args:
      easy_align: If True, the studs on the top of the brick will be capsules
        rather than cylinders. This makes alignment easier.
      flanges: Whether to use flanges on the bottom of the brick. These make the
        dynamics more expensive, but allow the bricks to be clicked together in
        partially overlapping configurations.
      variation: A float that controls the amount of variation in stud size (and
        therefore separation force). A value of 1.0 results in a distribution of
        separation forces that approximately matches the empirical distribution
        measured for real Duplo bricks. A value of 0.0 yields a deterministic
        separation force approximately equal to the mode of the empirical
        distribution.
      color: An optional tuple of (R, G, B) values specifying the color of the
        Duplo brick. These should be floats between 0 and 1. The default is red.

    Raises:
      ValueError: If `color` contains any value that is not between 0 and 1.
    """
    self._mjcf_root = mjcf.from_path(_DUPLO_XML_PATH)

    stud = self._mjcf_root.default.find('default', 'stud')
    if easy_align:
      # Make cylindrical studs invisible and disable contacts.
      stud.geom.group = 3
      stud.geom.contype = 9
      stud.geom.conaffinity = 8
      # Make capsule studs visible and enable contacts.
      stud_cap = self._mjcf_root.default.find('default', 'stud-capsule')
      stud_cap.geom.group = 0
      stud_cap.geom.contype = 0
      stud_cap.geom.conaffinity = 4
      self._active_stud_dclass = stud_cap
    else:
      self._active_stud_dclass = stud

    if flanges:
      flange_dclass = self._mjcf_root.default.find('default', 'flange')
      flange_dclass.geom.contype = 4  # Enable contact with flanges.

    stud_size = _STUD_SIZE_PARAMS[(easy_align, flanges)]
    offset = (1 - variation) * stud_size.lower_quartile
    self._lower = offset + variation * stud_size.minimum
    self._upper = offset + variation * stud_size.maximum

    self._studs = np.ndarray((2, 4), dtype=object)
    self._holes = np.ndarray((2, 4), dtype=object)

    for row in range(2):
      for column in range(4):
        self._studs[row, column] = self._mjcf_root.find(
            'site', 'stud_{}{}'.format(row, column))
        self._holes[row, column] = self._mjcf_root.find(
            'site', 'hole_{}{}'.format(row, column))

    if not all(0 <= value <= 1 for value in color):
      raise ValueError(_COLOR_NOT_BETWEEN_0_AND_1.format(color))
    self._mjcf_root.default.geom.rgba[:3] = color

  def initialize_episode_mjcf(self, random_state):
    """Randomizes the stud radius (and therefore the separation force)."""
    radius = random_state.uniform(self._lower, self._upper)
    self._active_stud_dclass.geom.size[0] = radius

  def _build_observables(self):
    return DuploObservables(self)

  @property
  def studs(self):
    """A (2, 4) numpy array of `mjcf.Elements` corresponding to stud sites."""
    return self._studs

  @property
  def holes(self):
    """A (2, 4) numpy array of `mjcf.Elements` corresponding to hole sites."""
    return self._holes

  @property
  def mjcf_model(self):
    return self._mjcf_root


class DuploObservables(composer.Observables, composer.FreePropObservableMixin):
  """Observables for the `Duplo` prop."""

  @define.observable
  def position(self):
    return observable.MJCFFeature(
        'sensordata',
        self._entity.mjcf_model.find('sensor', 'position'))

  @define.observable
  def orientation(self):
    return observable.MJCFFeature(
        'sensordata',
        self._entity.mjcf_model.find('sensor', 'orientation'))

  @define.observable
  def linear_velocity(self):
    return observable.MJCFFeature(
        'sensordata',
        self._entity.mjcf_model.find('sensor', 'linear_velocity'))

  @define.observable
  def angular_velocity(self):
    return observable.MJCFFeature(
        'sensordata',
        self._entity.mjcf_model.find('sensor', 'angular_velocity'))

  @define.observable
  def force(self):
    return observable.MJCFFeature(
        'sensordata',
        self._entity.mjcf_model.find('sensor', 'force'))



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Variations in colors.

Classes in this module allow users to specify a variations for each channel in
a variety of color spaces. The generated values are always RGBA arrays.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import colorsys

from dm_control.composer.variation import base
from dm_control.composer.variation import variation_values
import numpy as np


class RgbVariation(base.Variation):
  """Represents a variation in the RGB color space.

  This class allows users to specify independent variations in the R, G, B, and
  alpha channels of a color, and generates the corresponding array of RGBA
  values.
  """

  def __init__(self, r, g, b, alpha=1.0):
    self._r, self._g, self._b = r, g, b
    self._alpha = alpha

  def __call__(self, initial_value=None, current_value=None, random_state=None):
    return np.asarray(
        variation_values.evaluate([self._r, self._g, self._b, self._alpha],
                                  initial_value, current_value, random_state))


class HsvVariation(base.Variation):
  """Represents a variation in the HSV color space.

  This class allows users to specify independent variations in the H, S, V, and
  alpha channels of a color, and generates the corresponding array of RGBA
  values.
  """

  def __init__(self, h, s, v, alpha=1.0):
    self._h, self._s, self._v = h, s, v
    self._alpha = alpha

  def __call__(self, initial_value=None, current_value=None, random_state=None):
    h, s, v, alpha = variation_values.evaluate(
        (self._h, self._s, self._v, self._alpha), initial_value, current_value,
        random_state)
    return np.asarray(list(colorsys.hsv_to_rgb(h, s, v)) + [alpha])


class GrayVariation(HsvVariation):
  """Represents a variation in gray level.

  This class allows users to specify independent variations in the gray level
  and alpha channels of a color, and generates the corresponding array of RGBA
  values.
  """

  def __init__(self, gray_level, alpha=1.0):
    super(GrayVariation, self).__init__(h=0.0, s=0.0, v=gray_level, alpha=alpha)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Pendulum domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from dm_control.suite import base
from dm_control.suite import common
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np


_DEFAULT_TIME_LIMIT = 20
_ANGLE_BOUND = 8
_COSINE_BOUND = np.cos(np.deg2rad(_ANGLE_BOUND))
SUITE = containers.TaggedTasks()


def get_model_and_assets():
  """Returns a tuple containing the model XML string and a dict of assets."""
  return common.read_model('pendulum.xml'), common.ASSETS


@SUITE.add('benchmarking')
def swingup(time_limit=_DEFAULT_TIME_LIMIT, random=None,
            environment_kwargs=None):
  """Returns pendulum swingup task ."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = SwingUp(random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, time_limit=time_limit, **environment_kwargs)


class Physics(mujoco.Physics):
  """Physics simulation with additional features for the Pendulum domain."""

  def pole_vertical(self):
    """Returns vertical (z) component of pole frame."""
    return self.named.data.xmat['pole', 'zz']

  def angular_velocity(self):
    """Returns the angular velocity of the pole."""
    return self.named.data.qvel['hinge'].copy()

  def pole_orientation(self):
    """Returns both horizontal and vertical components of pole frame."""
    return self.named.data.xmat['pole', ['zz', 'xz']]


class SwingUp(base.Task):
  """A Pendulum `Task` to swing up and balance the pole."""

  def __init__(self, random=None):
    """Initialize an instance of `Pendulum`.

    Args:
      random: Optional, either a `numpy.random.RandomState` instance, an
        integer seed for creating a new `RandomState`, or None to select a seed
        automatically (default).
    """
    super(SwingUp, self).__init__(random=random)

  def initialize_episode(self, physics):
    """Sets the state of the environment at the start of each episode.

    Pole is set to a random angle between [-pi, pi).

    Args:
      physics: An instance of `Physics`.

    """
    physics.named.data.qpos['hinge'] = self.random.uniform(-np.pi, np.pi)
    super(SwingUp, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns an observation.

    Observations are states concatenating pole orientation and angular velocity
    and pixels from fixed camera.

    Args:
      physics: An instance of `physics`, Pendulum physics.

    Returns:
      A `dict` of observation.
    """
    obs = collections.OrderedDict()
    obs['orientation'] = physics.pole_orientation()
    obs['velocity'] = physics.angular_velocity()
    return obs

  def get_reward(self, physics):
    return rewards.tolerance(physics.pole_vertical(), (_COSINE_BOUND, 1))



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Parses MuJoCo header files and generates Python bindings."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import pprint
import textwrap

from absl import logging
from dm_control.autowrap import c_declarations
from dm_control.autowrap import codegen_util
from dm_control.autowrap import header_parsing
import pyparsing
import six

# Absolute path to the top-level module.
_MODULE = "dm_control.mujoco.wrapper"

# Imports used in all generated source files.
_BOILERPLATE_IMPORTS = [
    "from __future__ import absolute_import",
    "from __future__ import division",
    "from __future__ import print_function\n",
]


class Error(Exception):
  pass


class BindingGenerator(object):
  """Parses declarations from MuJoCo headers and generates Python bindings."""

  def __init__(self,
               enums_dict=None,
               consts_dict=None,
               typedefs_dict=None,
               hints_dict=None,
               types_dict=None,
               funcs_dict=None,
               strings_dict=None,
               func_ptrs_dict=None,
               index_dict=None):
    """Constructs a new HeaderParser instance.

    The optional arguments listed below can be used to passing in dict-like
    objects specifying pre-defined declarations. By default empty
    UniqueOrderedDicts will be instantiated and then populated according to the
    contents of the headers.

    Args:
      enums_dict: Nested mappings from {enum_name: {member_name: value}}.
      consts_dict: Mapping from {const_name: value}.
      typedefs_dict: Mapping from {type_name: ctypes_typename}.
      hints_dict: Mapping from {var_name: shape_tuple}.
      types_dict: Mapping from {type_name: type_instance}.
      funcs_dict: Mapping from {func_name: Function_instance}.
      strings_dict: Mapping from {var_name: StaticStringArray_instance}.
      func_ptrs_dict: Mapping from {var_name: FunctionPtr_instance}.
      index_dict: Mapping from {lowercase_struct_name: {var_name: shape_tuple}}.
    """
    self.enums_dict = (enums_dict if enums_dict is not None
                       else codegen_util.UniqueOrderedDict())
    self.consts_dict = (consts_dict if consts_dict is not None
                        else codegen_util.UniqueOrderedDict())
    self.typedefs_dict = (typedefs_dict if typedefs_dict is not None
                          else codegen_util.UniqueOrderedDict())
    self.hints_dict = (hints_dict if hints_dict is not None
                       else codegen_util.UniqueOrderedDict())
    self.types_dict = (types_dict if types_dict is not None
                       else codegen_util.UniqueOrderedDict())
    self.funcs_dict = (funcs_dict if funcs_dict is not None
                       else codegen_util.UniqueOrderedDict())
    self.strings_dict = (strings_dict if strings_dict is not None
                         else codegen_util.UniqueOrderedDict())
    self.func_ptrs_dict = (func_ptrs_dict if func_ptrs_dict is not None
                           else codegen_util.UniqueOrderedDict())
    self.index_dict = (index_dict if index_dict is not None
                       else codegen_util.UniqueOrderedDict())

  def get_consts_and_enums(self):
    consts_and_enums = self.consts_dict.copy()
    for enum in six.itervalues(self.enums_dict):
      consts_and_enums.update(enum)
    return consts_and_enums

  def resolve_size(self, old_size):
    """Resolves an array size identifier.

    The following conversions will be attempted:

      * If `old_size` is an integer it will be returned as-is.
      * If `old_size` is a string of the form `"3"` it will be cast to an int.
      * If `old_size` is a string in `self.consts_dict` then the value of the
        constant will be returned.
      * If `old_size` is a string of the form `"3*constant_name"` then the
        result of `3*constant_value` will be returned.
      * If `old_size` is a string that does not specify an int constant and
        cannot be cast to an int (e.g. an identifier for a dynamic dimension,
        such as `"ncontact"`) then it will be returned as-is.

    Args:
      old_size: An int or string.

    Returns:
      An int or string.
    """
    if isinstance(old_size, int):
      return old_size  # If it's already an int then there's nothing left to do
    elif "*" in old_size:
      # If it's a string specifying a product (such as "2*mjMAXLINEPNT"),
      # recursively resolve the components to ints and calculate the result.
      size = 1
      for part in old_size.split("*"):
        dim = self.resolve_size(part)
        assert isinstance(dim, int)
        size *= dim
      return size
    else:
      # Recursively dereference any sizes declared in header macros
      size = codegen_util.recursive_dict_lookup(old_size,
                                                self.get_consts_and_enums())
      # Try to coerce the result to an int, return a string if this fails
      return codegen_util.try_coerce_to_num(size, try_types=(int,))

  def get_shape_tuple(self, old_size, squeeze=False):
    """Generates a shape tuple from parser results.

    Args:
      old_size: Either a `pyparsing.ParseResults`, or a valid int or string
       input to `self.resolve_size` (see method docstring for further details).
      squeeze: If True, any dimensions that are statically defined as 1 will be
        removed from the shape tuple.

    Returns:
      A shape tuple containing ints for dimensions that are statically defined,
      and string size identifiers for dimensions that can only be determined at
      runtime.
    """
    if isinstance(old_size, pyparsing.ParseResults):
      # For multi-dimensional arrays, convert each dimension separately
      shape = tuple(self.resolve_size(dim) for dim in old_size)
    else:
      shape = (self.resolve_size(old_size),)
    if squeeze:
      shape = tuple(d for d in shape if d != 1)  # Remove singleton dimensions
    return shape

  def resolve_typename(self, old_ctypes_typename):
    """Gets a qualified ctypes typename from typedefs_dict and C_TO_CTYPES."""

    # Recursively dereference any typenames declared in self.typedefs_dict
    new_ctypes_typename = codegen_util.recursive_dict_lookup(
        old_ctypes_typename, self.typedefs_dict)

    # Try to convert to a ctypes native typename
    new_ctypes_typename = header_parsing.C_TO_CTYPES.get(
        new_ctypes_typename, new_ctypes_typename)

    if new_ctypes_typename == old_ctypes_typename:
      logging.warning("Could not resolve typename '%s'", old_ctypes_typename)

    return new_ctypes_typename

  def get_type_from_token(self, token, parent=None):
    """Accepts a token returned by a parser, returns a subclass of CDeclBase."""

    comment = codegen_util.mangle_comment(token.comment)
    is_const = token.is_const == "const"

    # An anonymous union declaration
    if token.anonymous_union:
      if not parent and parent.name:
        raise Error(
            "Anonymous unions must be members of a named struct or union.")

      # Generate a name based on the name of the parent.
      name = codegen_util.mangle_varname(parent.name + "_anon_union")

      members = codegen_util.UniqueOrderedDict()
      sub_structs = codegen_util.UniqueOrderedDict()
      out = c_declarations.AnonymousUnion(
          name, members, sub_structs, comment, parent)

      # Add members
      for sub_token in token.members:

        # Recurse into nested structs
        member = self.get_type_from_token(sub_token, parent=out)
        out.members[member.name] = member

        # Nested sub-structures need special treatment
        if isinstance(member, c_declarations.Struct):
          out.sub_structs[member.name] = member

      # Add to dict of unions
      self.types_dict[out.ctypes_typename] = out

    # A struct declaration
    elif token.members:

      name = token.name

      # If the name is empty, see if there is a type declaration that matches
      # this struct's typename
      if not name:
        for k, v in six.iteritems(self.typedefs_dict):
          if v == token.typename:
            name = k

      # Anonymous structs need a dummy typename
      typename = token.typename
      if not typename:
        if parent:
          typename = token.name
        else:
          raise Error(
              "Anonymous structs that aren't members of a named struct are not "
              "supported (name = '{token.name}').".format(token=token))

      # Mangle the name if it contains any protected keywords
      name = codegen_util.mangle_varname(name)

      members = codegen_util.UniqueOrderedDict()
      sub_structs = codegen_util.UniqueOrderedDict()
      out = c_declarations.Struct(name, typename, members, sub_structs, comment,
                                  parent, is_const)

      # Map the old typename to the mangled typename in typedefs_dict
      self.typedefs_dict[typename] = out.ctypes_typename

      # Add members
      for sub_token in token.members:

        # Recurse into nested structs
        member = self.get_type_from_token(sub_token, parent=out)
        out.members[member.name] = member

        # Nested sub-structures need special treatment
        if isinstance(member, c_declarations.Struct):
          out.sub_structs[member.name] = member

      # Add to dict of structs
      self.types_dict[out.ctypes_typename] = out

    else:

      name = codegen_util.mangle_varname(token.name)
      typename = self.resolve_typename(token.typename)

      # 1D array with size defined at compile time
      if token.size:
        shape = self.get_shape_tuple(token.size)
        if typename in {header_parsing.NONE, header_parsing.CTYPES_CHAR}:
          out = c_declarations.StaticPtrArray(
              name, typename, shape, comment, parent, is_const)
        else:
          out = c_declarations.StaticNDArray(
              name, typename, shape, comment, parent, is_const)

      elif token.ptr:

        # Pointer to a numpy-compatible type, could be an array or a scalar
        if typename in header_parsing.CTYPES_TO_NUMPY:

          # Multidimensional array (one or more dimensions might be undefined)
          if name in self.hints_dict:

            # Dynamically-sized dimensions have string identifiers
            shape = self.hints_dict[name]
            if any(isinstance(d, six.string_types) for d in shape):
              out = c_declarations.DynamicNDArray(name, typename, shape,
                                                  comment, parent, is_const)
            else:
              out = c_declarations.StaticNDArray(name, typename, shape, comment,
                                                 parent, is_const)

          # This must be a pointer to a scalar primitive
          else:
            out = c_declarations.ScalarPrimitivePtr(name, typename, comment,
                                                    parent, is_const)

        # Pointer to struct or other arbitrary type
        else:
          out = c_declarations.ScalarPrimitivePtr(name, typename, comment,
                                                  parent, is_const)

      # A struct we've already encountered
      elif typename in self.types_dict:
        s = self.types_dict[typename]
        if isinstance(s, c_declarations.FunctionPtrTypedef):
          out = c_declarations.FunctionPtr(
              name, token.name, s.typename, comment)
        else:
          out = c_declarations.Struct(name, s.typename, s.members,
                                      s.sub_structs, comment, parent)

      # Presumably this is a scalar primitive
      else:
        out = c_declarations.ScalarPrimitive(name, typename, comment, parent,
                                             is_const)

    return out

  # Parsing functions.
  # ----------------------------------------------------------------------------

  def parse_hints(self, xmacro_src):
    """Parses mjxmacro.h, update self.hints_dict."""
    parser = header_parsing.XMACRO
    for tokens, _, _ in parser.scanString(xmacro_src):
      for xmacro in tokens:
        for member in xmacro.members:
          # "Squeeze out" singleton dimensions.
          shape = self.get_shape_tuple(member.dims, squeeze=True)
          self.hints_dict.update({member.name: shape})

          if codegen_util.is_macro_pointer(xmacro.name):
            struct_name = codegen_util.macro_struct_name(xmacro.name)
            if struct_name not in self.index_dict:
              self.index_dict[struct_name] = {}

            self.index_dict[struct_name].update({member.name: shape})

  def parse_enums(self, src):
    """Parses mj*.h, update self.enums_dict."""
    parser = header_parsing.ENUM_DECL
    for tokens, _, _ in parser.scanString(src):
      for enum in tokens:
        members = codegen_util.UniqueOrderedDict()
        value = 0
        for member in enum.members:
          # Leftward bitshift
          if member.bit_lshift_a:
            value = int(member.bit_lshift_a) << int(member.bit_lshift_b)
          # Assignment
          elif member.value:
            value = int(member.value)
          # Implicit count
          else:
            value += 1
          members.update({member.name: value})
        self.enums_dict.update({enum.name: members})

  def parse_consts_typedefs(self, src):
    """Updates self.consts_dict, self.typedefs_dict."""
    parser = (header_parsing.COND_DECL |
              header_parsing.UNCOND_DECL)
    for tokens, _, _ in parser.scanString(src):
      self.recurse_into_conditionals(tokens)

  def recurse_into_conditionals(self, tokens):
    """Called recursively within nested #if(n)def... #else... #endif blocks."""
    for token in tokens:
      # Another nested conditional block
      if token.predicate:
        if (token.predicate in self.get_consts_and_enums()
            and self.get_consts_and_enums()[token.predicate]):
          self.recurse_into_conditionals(token.if_true)
        else:
          self.recurse_into_conditionals(token.if_false)
      # One or more declarations
      else:
        if token.typename:
          self.typedefs_dict.update({token.name: token.typename})
        elif token.value:
          value = codegen_util.try_coerce_to_num(token.value)
          # Avoid adding function aliases.
          if isinstance(value, six.string_types):
            continue
          else:
            self.consts_dict.update({token.name: value})
        else:
          self.consts_dict.update({token.name: True})

  def parse_structs_and_function_pointer_typedefs(self, src):
    """Updates self.types_dict."""
    parser = (header_parsing.NESTED_STRUCTS |
              header_parsing.FUNCTION_PTR_TYPE_DECL)
    for tokens, _, _ in parser.scanString(src):
      for token in tokens:
        if token.return_type:
          # This is a function type declaration.
          self.types_dict[token.typename] = c_declarations.FunctionPtrTypedef(
              token.typename,
              self.get_type_from_token(token.return_type),
              tuple(self.get_type_from_token(arg) for arg in token.arguments))
        else:
          # This is a struct or a union.
          self.get_type_from_token(token)

  def parse_functions(self, src):
    """Updates self.funcs_dict."""
    parser = header_parsing.MJAPI_FUNCTION_DECL
    for tokens, _, _ in parser.scanString(src):
      for token in tokens:
        name = codegen_util.mangle_varname(token.name)
        comment = codegen_util.mangle_comment(token.comment)
        if token.arguments:
          args = codegen_util.UniqueOrderedDict()
          for arg in token.arguments:
            a = self.get_type_from_token(arg)
            args[a.name] = a
        else:
          args = None
        if token.return_value:
          ret_val = self.get_type_from_token(token.return_value)
        else:
          ret_val = None
        func = c_declarations.Function(name, args, ret_val, comment)
        self.funcs_dict[func.name] = func

  def parse_global_strings(self, src):
    """Updates self.strings_dict."""
    parser = header_parsing.MJAPI_STRING_ARRAY
    for token, _, _ in parser.scanString(src):
      name = codegen_util.mangle_varname(token.name)
      shape = self.get_shape_tuple(token.dims)
      self.strings_dict[name] = c_declarations.StaticStringArray(
          name, shape, symbol_name=token.name)

  def parse_function_pointers(self, src):
    """Updates self.func_ptrs_dict."""
    parser = header_parsing.MJAPI_FUNCTION_PTR
    for token, _, _ in parser.scanString(src):
      name = codegen_util.mangle_varname(token.name)
      self.func_ptrs_dict[name] = c_declarations.FunctionPtr(
          name, symbol_name=token.name,
          type_name=token.typename, comment=token.comment)

  # Code generation methods
  # ----------------------------------------------------------------------------

  def make_header(self, imports=()):
    """Returns a header string for an auto-generated Python source file."""
    docstring = textwrap.dedent("""
    \"\"\"Automatically generated by {scriptname:}.

    MuJoCo header version: {mujoco_version:}
    \"\"\"
    """.format(scriptname=os.path.split(__file__)[-1],
               mujoco_version=self.consts_dict["mjVERSION_HEADER"]))
    docstring = docstring[1:]  # Strip the leading line break.
    return "\n".join(
        [docstring] + _BOILERPLATE_IMPORTS + list(imports) + ["\n"])

  def write_consts(self, fname):
    """Write constants."""
    imports = [
        "# pylint: disable=invalid-name",
    ]
    with open(fname, "w") as f:
      f.write(self.make_header(imports))
      f.write(codegen_util.comment_line("Constants") + "\n")
      for name, value in six.iteritems(self.consts_dict):
        f.write("{0} = {1}\n".format(name, value))
      f.write("\n" + codegen_util.comment_line("End of generated code"))

  def write_enums(self, fname):
    """Write enum definitions."""
    with open(fname, "w") as f:
      imports = [
          "import collections",
          "# pylint: disable=invalid-name",
          "# pylint: disable=line-too-long",
      ]
      f.write(self.make_header(imports))
      f.write(codegen_util.comment_line("Enums"))
      for enum_name, members in six.iteritems(self.enums_dict):
        fields = ["\"{}\"".format(name) for name in six.iterkeys(members)]
        values = [str(value) for value in six.itervalues(members)]
        s = textwrap.dedent("""
        {0} = collections.namedtuple(
            "{0}",
            [{1}]
        )({2})
        """).format(enum_name, ",\n     ".join(fields), ", ".join(values))
        f.write(s)
      f.write("\n" + codegen_util.comment_line("End of generated code"))

  def write_types(self, fname):
    """Write ctypes struct and function type declarations."""
    imports = [
        "import ctypes",
    ]
    with open(fname, "w") as f:
      f.write(self.make_header(imports))
      f.write(codegen_util.comment_line(
          "ctypes struct, union, and function type declarations"))
      for type_decl in six.itervalues(self.types_dict):
        f.write("\n" + type_decl.ctypes_decl)
      f.write("\n" + codegen_util.comment_line("End of generated code"))

  def write_wrappers(self, fname):
    """Write wrapper classes for ctypes structs."""
    with open(fname, "w") as f:
      imports = [
          "import ctypes",
          "# pylint: disable=undefined-variable",
          "# pylint: disable=wildcard-import",
          "from {} import util".format(_MODULE),
          "from {}.mjbindings.types import *".format(_MODULE),
      ]
      f.write(self.make_header(imports))
      f.write(codegen_util.comment_line("Low-level wrapper classes"))
      for type_decl in six.itervalues(self.types_dict):
        if isinstance(type_decl, c_declarations.Struct):
          f.write("\n" + type_decl.wrapper_class)
      f.write("\n" + codegen_util.comment_line("End of generated code"))

  def write_funcs_and_globals(self, fname):
    """Write ctypes declarations for functions and global data."""
    imports = [
        "import collections",
        "import ctypes",
        "# pylint: disable=undefined-variable",
        "# pylint: disable=wildcard-import",
        "from {} import util".format(_MODULE),
        "from {}.mjbindings.types import *".format(_MODULE),
        "import numpy as np",
        "# pylint: disable=line-too-long",
        "# pylint: disable=invalid-name",
        "# common_typos_disable",
    ]
    with open(fname, "w") as f:
      f.write(self.make_header(imports))
      f.write("mjlib = util.get_mjlib()\n")

      f.write("\n" + codegen_util.comment_line("ctypes function declarations"))
      for function in six.itervalues(self.funcs_dict):
        f.write("\n" + function.ctypes_func_decl(cdll_name="mjlib"))

      # Only require strings for UI purposes.
      f.write("\n" + codegen_util.comment_line("String arrays") + "\n")
      for string_arr in six.itervalues(self.strings_dict):
        f.write(string_arr.ctypes_var_decl(cdll_name="mjlib"))

      f.write("\n" + codegen_util.comment_line("Callback function pointers"))

      fields = ["'_{0}'".format(func_ptr.name)
                for func_ptr in self.func_ptrs_dict.values()]
      values = [func_ptr.ctypes_var_decl(cdll_name="mjlib")
                for func_ptr in self.func_ptrs_dict.values()]
      f.write(textwrap.dedent("""
        class _Callbacks(object):

          __slots__ = [
              {0}
          ]

          def __init__(self):
            {1}
        """).format(",\n      ".join(fields), "\n    ".join(values)))

      indent = codegen_util.Indenter()
      with indent:
        for func_ptr in self.func_ptrs_dict.values():
          f.write(indent(func_ptr.getters_setters_with_custom_prefix("self._")))

      f.write("\n\ncallbacks = _Callbacks()  # pylint: disable=invalid-name")
      f.write("\ndel _Callbacks\n")

      f.write("\n" + codegen_util.comment_line("End of generated code"))

  def write_index_dict(self, fname):
    """Write file containing array shape information for indexing."""
    pp = pprint.PrettyPrinter()
    output_string = pp.pformat(dict(self.index_dict))
    indent = codegen_util.Indenter()
    imports = [
        "# pylint: disable=bad-continuation",
        "# pylint: disable=line-too-long",
    ]
    with open(fname, "w") as f:
      f.write(self.make_header(imports))
      f.write("array_sizes = (\n")
      with indent:
        f.write(output_string)
      f.write("\n)")
      f.write("\n" + codegen_util.comment_line("End of generated code"))



# Copyright 2018-2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Module containing abstract base classes for Composer environments."""

from dm_control.composer.arena import Arena
from dm_control.composer.constants import *  # pylint: disable=wildcard-import
from dm_control.composer.define import cached_property
from dm_control.composer.define import observable
from dm_control.composer.entity import Entity
from dm_control.composer.entity import FreePropObservableMixin
from dm_control.composer.entity import ModelWrapperEntity
from dm_control.composer.entity import Observables
from dm_control.composer.environment import Environment
from dm_control.composer.environment import EpisodeInitializationError
from dm_control.composer.environment import HOOK_NAMES
from dm_control.composer.initializer import Initializer
from dm_control.composer.robot import Robot
from dm_control.composer.task import NullTask
from dm_control.composer.task import Task



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests of the renderer module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized
from dm_control.mujoco import wrapper
from dm_control.mujoco.wrapper.mjbindings import enums
from dm_control.mujoco.wrapper.mjbindings import types
from dm_control.viewer import renderer
import mock
import numpy as np


renderer.mjlib = mock.MagicMock()

_SCREEN_SIZE = types.MJRRECT(0, 0, 320, 240)


class BaseRendererTest(absltest.TestCase):

  class MockRenderer(renderer.BaseRenderer):
    pass

  class MockRenderComponent(renderer.Component):

    counter = 0

    def __init__(self):
      self._call_order = -1

    def render(self, context, viewport):
      self._call_order = BaseRendererTest.MockRenderComponent.counter
      BaseRendererTest.MockRenderComponent.counter += 1

    @property
    def call_order(self):
      return self._call_order

  def setUp(self):
    super(BaseRendererTest, self).setUp()
    self.renderer = BaseRendererTest.MockRenderer()
    self.context = mock.MagicMock()
    self.viewport = mock.MagicMock()

  def test_rendering_components(self):
    regular_component = BaseRendererTest.MockRenderComponent()
    screen_capture_component = BaseRendererTest.MockRenderComponent()
    self.renderer.components += regular_component
    self.renderer.screen_capture_components += screen_capture_component
    self.renderer._render_components(self.context, self.viewport)
    self.assertEqual(0, regular_component.call_order)
    self.assertEqual(1, screen_capture_component.call_order)


class OffScreenRendererTest(absltest.TestCase):

  def setUp(self):
    super(OffScreenRendererTest, self).setUp()
    self.model = mock.MagicMock()
    self.model.vis.global_.offwidth = _SCREEN_SIZE.width
    self.model.vis.global_.offheight = _SCREEN_SIZE.height

    self.surface = mock.MagicMock()
    self.renderer = renderer.OffScreenRenderer(self.model, self.surface)
    self.renderer._mujoco_context = mock.MagicMock()

    self.viewport = mock.MagicMock()
    self.scene = mock.MagicMock()

    self.viewport.width = 3
    self.viewport.height = 3
    self.viewport.dimensions = np.array([3, 3])

  def test_render_context_initialization(self):
    self.renderer._mujoco_context = None
    self.renderer.render(self.viewport, self.scene)
    self.assertIsNotNone(self.renderer._mujoco_context)

  def test_resizing_pixel_buffer_to_viewport_size(self):
    self.renderer.render(self.viewport, self.scene)
    self.assertEqual((self.viewport.width, self.viewport.height, 3),
                     self.renderer._rgb_buffer.shape)

  def test_rendering_components(self):
    regular_component = mock.MagicMock()
    screen_capture_components = mock.MagicMock()
    self.renderer.components += [regular_component]
    self.renderer.screen_capture_components += [screen_capture_components]
    self.renderer._render_on_gl_thread(self.viewport, self.scene)
    regular_component.render.assert_called_once()
    screen_capture_components.render.assert_called_once()


class PerturbationTest(absltest.TestCase):

  def setUp(self):
    super(PerturbationTest, self).setUp()
    self.model = mock.MagicMock()
    self.data = mock.MagicMock()
    self.scene = mock.MagicMock()
    self.valid_pos = np.array([1, 2, 3])

    self.body_id = 0
    self.data.xpos = [np.array([0, 1, 2])]
    self.data.xmat = [np.identity(3)]

    self.perturbation = renderer.Perturbation(
        self.body_id, self.model, self.data, self.scene)

    renderer.mjlib.reset_mock()

  def test_start_params_validation(self):
    self.perturbation.start_move(None, self.valid_pos)
    self.assertEqual(0, renderer.mjlib.mjv_initPerturb.call_count)
    self.assertEqual(enums.mjtMouse.mjMOUSE_NONE, self.perturbation._action)

    self.perturbation.start_move(enums.mjtMouse.mjMOUSE_MOVE_V, None)
    self.assertEqual(0, renderer.mjlib.mjv_initPerturb.call_count)
    self.assertEqual(enums.mjtMouse.mjMOUSE_NONE, self.perturbation._action)

  def test_starting_an_operation(self):
    self.perturbation.start_move(enums.mjtMouse.mjMOUSE_MOVE_V, self.valid_pos)
    renderer.mjlib.mjv_initPerturb.assert_called_once()
    self.assertEqual(enums.mjtMouse.mjMOUSE_MOVE_V, self.perturbation._action)

  def test_starting_translation(self):
    self.perturbation.start_move(enums.mjtMouse.mjMOUSE_MOVE_V, self.valid_pos)
    self.assertEqual(
        enums.mjtPertBit.mjPERT_TRANSLATE, self.perturbation._perturb.active)

  def test_starting_rotation(self):
    self.perturbation.start_move(enums.mjtMouse.mjMOUSE_ROTATE_V,
                                 self.valid_pos)
    self.assertEqual(
        enums.mjtPertBit.mjPERT_ROTATE, self.perturbation._perturb.active)

  def test_starting_grip_transform(self):
    self.perturbation.start_move(enums.mjtMouse.mjMOUSE_MOVE_V, self.valid_pos)
    np.testing.assert_array_equal(
        [1, 1, 1], self.perturbation._perturb.localpos)

  def test_ticking_operation(self):
    self.perturbation._action = enums.mjtMouse.mjMOUSE_MOVE_V
    self.perturbation.tick_move([.1, .2])
    renderer.mjlib.mjv_movePerturb.assert_called_once()
    action, dx, dy = renderer.mjlib.mjv_movePerturb.call_args[0][2:5]
    self.assertEqual(self.perturbation._action, action)
    self.assertEqual(.1, dx)
    self.assertEqual(.2, dy)

  def test_ticking_stopped_operation_yields_no_results(self):
    self.perturbation._action = None
    self.perturbation.tick_move([.1, .2])
    self.assertEqual(0, renderer.mjlib.mjv_movePerturb.call_count)

    self.perturbation._action = enums.mjtMouse.mjMOUSE_NONE
    self.perturbation.tick_move([.1, .2])
    self.assertEqual(0, renderer.mjlib.mjv_movePerturb.call_count)

  def test_stopping_operation(self):
    self.perturbation._action = enums.mjtMouse.mjMOUSE_MOVE_V
    self.perturbation._perturb.active = enums.mjtPertBit.mjPERT_TRANSLATE
    self.perturbation.end_move()
    self.assertEqual(enums.mjtMouse.mjMOUSE_NONE, self.perturbation._action)
    self.assertEqual(0, self.perturbation._perturb.active)

  def test_applying_operation_results_while_not_paused(self):
    with self.perturbation.apply(False):
      renderer.mjlib.mjv_applyPerturbPose.assert_called_once()
      self.assertEqual(0, renderer.mjlib.mjv_applyPerturbPose.call_args[0][3])
      renderer.mjlib.mjv_applyPerturbForce.assert_called_once()

  def test_applying_operation_results_while_paused(self):
    with self.perturbation.apply(True):
      renderer.mjlib.mjv_applyPerturbPose.assert_called_once()
      self.assertEqual(1, renderer.mjlib.mjv_applyPerturbPose.call_args[0][3])
      self.assertEqual(0, renderer.mjlib.mjv_applyPerturbForce.call_count)

  def test_clearing_applied_forces_after_appling_operation(self):
    self.data.xfrc_applied = np.zeros(1)
    with self.perturbation.apply(True):
      # At this point the simulation will calculate forces to apply and assign
      # them to a proper MjvData structure field, as we're doing below.
      self.data.xfrc_applied[self.body_id] = 1

    # While exiting, the context clears that information.
    self.assertEqual(0, self.data.xfrc_applied[self.body_id])


class RenderSettingsTest(absltest.TestCase):

  def setUp(self):
    super(RenderSettingsTest, self).setUp()
    self.settings = renderer.RenderSettings()
    self.scene = wrapper.MjvScene()

  def test_applying_settings(self):
    self.settings._stereo_mode = 5
    self.settings._render_flags[:] = np.arange(len(self.settings._render_flags))
    self.settings.apply_settings(self.scene)
    self.assertEqual(self.settings._stereo_mode, self.scene.stereo)
    np.testing.assert_array_equal(self.settings._render_flags, self.scene.flags)

  def test_toggle_rendering_flag(self):
    self.settings._render_flags[0] = 1
    self.settings.toggle_rendering_flag(0)
    self.assertEqual(0, self.settings._render_flags[0])
    self.settings.toggle_rendering_flag(0)
    self.assertEqual(1, self.settings._render_flags[0])

  def test_toggle_visualization_flag(self):
    self.settings._visualization_options.flags[0] = 1
    self.settings.toggle_visualization_flag(0)
    self.assertEqual(0, self.settings._visualization_options.flags[0])
    self.settings.toggle_visualization_flag(0)
    self.assertEqual(1, self.settings._visualization_options.flags[0])

  def test_toggle_geom_group(self):
    self.settings._visualization_options.geomgroup[0] = 1
    self.settings.toggle_geom_group(0)
    self.assertEqual(0, self.settings._visualization_options.geomgroup[0])
    self.settings.toggle_geom_group(0)
    self.assertEqual(1, self.settings._visualization_options.geomgroup[0])

  def test_toggle_site_group(self):
    self.settings._visualization_options.sitegroup[0] = 1
    self.settings.toggle_site_group(0)
    self.assertEqual(0, self.settings._visualization_options.sitegroup[0])
    self.settings.toggle_site_group(0)
    self.assertEqual(1, self.settings._visualization_options.sitegroup[0])

  def test_toggle_stereo_buffering(self):
    self.settings.toggle_stereo_buffering()
    self.assertEqual(enums.mjtStereo.mjSTEREO_QUADBUFFERED,
                     self.settings._stereo_mode)
    self.settings.toggle_stereo_buffering()
    self.assertEqual(enums.mjtStereo.mjSTEREO_NONE,
                     self.settings._stereo_mode)

  def test_cycling_forward_through_render_modes(self):
    self.settings._visualization_options.frame = 0
    self.settings.select_next_rendering_mode()
    self.assertEqual(1, self.settings._visualization_options.frame)

    self.settings._visualization_options.frame = enums.mjtFrame.mjNFRAME - 1
    self.settings.select_next_rendering_mode()
    self.assertEqual(0, self.settings._visualization_options.frame)

  def test_cycling_backward_through_render_modes(self):
    self.settings._visualization_options.frame = 0
    self.settings.select_prev_rendering_mode()
    self.assertEqual(enums.mjtFrame.mjNFRAME - 1,
                     self.settings._visualization_options.frame)

    self.settings._visualization_options.frame = 1
    self.settings.select_prev_rendering_mode()
    self.assertEqual(0, self.settings._visualization_options.frame)

  def test_cycling_forward_through_labeling_modes(self):
    self.settings._visualization_options.label = 0
    self.settings.select_next_labeling_mode()
    self.assertEqual(1, self.settings._visualization_options.label)

    self.settings._visualization_options.label = enums.mjtLabel.mjNLABEL - 1
    self.settings.select_next_labeling_mode()
    self.assertEqual(0, self.settings._visualization_options.label)

  def test_cycling_backward_through_labeling_modes(self):
    self.settings._visualization_options.label = 0
    self.settings.select_prev_labeling_mode()
    self.assertEqual(enums.mjtLabel.mjNLABEL - 1,
                     self.settings._visualization_options.label)

    self.settings._visualization_options.label = 1
    self.settings.select_prev_labeling_mode()
    self.assertEqual(0, self.settings._visualization_options.label)


class SceneCameraTest(parameterized.TestCase):

  def setUp(self):
    super(SceneCameraTest, self).setUp()
    self.model = mock.MagicMock()
    self.data = mock.MagicMock()
    self.options = mock.MagicMock()
    with mock.patch.object(renderer.wrapper.core.mjlib,
                           'mjv_makeScene') as mock_make_scene:
      self.camera = renderer.SceneCamera(self.model, self.data, self.options)
    mock_make_scene.assert_called_once()

  def test_freelook_mode(self):
    self.camera.set_freelook_mode()
    self.assertEqual(-1, self.camera._camera.trackbodyid)
    self.assertEqual(-1, self.camera._camera.fixedcamid)
    self.assertEqual(enums.mjtCamera.mjCAMERA_FREE, self.camera._camera.type_)
    self.assertEqual('Free', self.camera.name)

  def test_tracking_mode(self):
    body_id = 5
    self.camera.set_tracking_mode(body_id)
    self.assertEqual(body_id, self.camera._camera.trackbodyid)
    self.assertEqual(-1, self.camera._camera.fixedcamid)
    self.assertEqual(enums.mjtCamera.mjCAMERA_TRACKING,
                     self.camera._camera.type_)

    self.model.id2name = mock.MagicMock(return_value='body_name')
    self.assertEqual('Tracking body "body_name"', self.camera.name)

  def test_fixed_mode(self):
    camera_id = 5
    self.camera.set_fixed_mode(camera_id)
    self.assertEqual(-1, self.camera._camera.trackbodyid)
    self.assertEqual(camera_id, self.camera._camera.fixedcamid)
    self.assertEqual(enums.mjtCamera.mjCAMERA_FIXED,
                     self.camera._camera.type_)

    self.model.id2name = mock.MagicMock(return_value='camera_name')
    self.assertEqual('camera_name', self.camera.name)

  def test_look_at(self):
    target_pos = [10, 20, 30]
    distance = 5.
    self.camera.look_at(target_pos, distance)
    np.testing.assert_array_equal(target_pos, self.camera._camera.lookat)
    np.testing.assert_array_equal(distance, self.camera._camera.distance)

  def test_moving_camera(self):
    action = enums.mjtMouse.mjMOUSE_MOVE_V
    offset = [0.1, -0.2]
    with mock.patch(renderer.__name__ + '.mjlib') as mock_mjlib:
      self.camera.move(action, offset)
      mock_mjlib.mjv_moveCamera.assert_called_once()

  def test_zoom_to_scene(self):
    scene_center = np.array([1, 2, 3])
    scene_extents = np.array([10, 20, 30])

    self.camera.look_at = mock.MagicMock()
    self.model.stat = mock.MagicMock()
    self.model.stat.center = scene_center
    self.model.stat.extent = scene_extents

    self.camera.zoom_to_scene()
    self.camera.look_at.assert_called_once()
    np.testing.assert_array_equal(
        scene_center, self.camera.look_at.call_args[0][0])
    np.testing.assert_array_equal(
        scene_extents * 1.5, self.camera.look_at.call_args[0][1])

  def test_camera_transform(self):
    self.camera._scene.camera[0].up[:] = [0, 1, 0]
    self.camera._scene.camera[0].forward[:] = [0, 0, 1]
    self.camera._scene.camera[0].pos[:] = [5, 0, 0]
    self.camera._scene.camera[1].pos[:] = [10, 0, 0]

    rotation_mtx, position = self.camera.transform
    np.testing.assert_array_equal([-1, 0, 0], rotation_mtx[0])
    np.testing.assert_array_equal([0, 1, 0], rotation_mtx[1])
    np.testing.assert_array_equal([0, 0, 1], rotation_mtx[2])
    np.testing.assert_array_equal([7.5, 0, 0], position)

  @parameterized.parameters(
      (0, 0, False),
      (0, 1, False),
      (1, 0, False),
      (2, 1, False),
      (1, 2, True))
  def test_is_camera_initialized(self, frustum_near, frustum_far, result):
    gl_camera = mock.MagicMock()
    self.camera._scene = mock.MagicMock()
    self.camera._scene.ptr.contents.camera = [gl_camera]

    gl_camera.frustum_near = frustum_near
    gl_camera.frustum_far = frustum_far
    self.assertEqual(result, self.camera.is_initialized)


class RaycastsTest(absltest.TestCase):

  def setUp(self):
    super(RaycastsTest, self).setUp()
    self.model = mock.MagicMock()
    self.data = mock.MagicMock()
    self.options = mock.MagicMock()

    self.viewport = mock.MagicMock()
    with mock.patch.object(renderer.wrapper.core.mjlib,
                           'mjv_makeScene') as mock_make_scene:
      self.camera = renderer.SceneCamera(self.model, self.data, self.options)
    mock_make_scene.assert_called_once()
    self.initialize_camera(True)

  def initialize_camera(self, enable):
    gl_camera = mock.MagicMock()
    self.camera._scene = mock.MagicMock()
    self.camera._scene.ptr.contents.camera = [gl_camera]
    gl_camera.frustum_near = 1 if enable else 0
    gl_camera.frustum_far = 2 if enable else 0

  def test_raycast_mapping_geom_to_body_id(self):
    def build_mjv_select(mock_body_id, mock_geom_id, mock_position):
      def mock_select(
          m, d, vopt, aspectratio, relx, rely, scn, selpnt, geomid, skinid):
        del m, d, vopt, aspectratio, relx, rely, scn, skinid  # Unused.
        selpnt[:] = mock_position
        geomid[:] = mock_geom_id
        return mock_body_id
      return mock_select

    geom_id = 0
    body_id = 5
    world_pos = [1, 2, 3]
    self.model.geom_bodyid = np.zeros(10)
    self.model.geom_bodyid[geom_id] = body_id
    mock_select = build_mjv_select(body_id, geom_id, world_pos)

    with mock.patch(renderer.__name__ + '.mjlib') as mock_mjlib:
      mock_mjlib.mjv_select = mock.MagicMock(side_effect=mock_select)
      hit_body_id, hit_world_pos = self.camera.raycast(self.viewport, [0, 0])
      self.assertEqual(hit_body_id, body_id)
      np.testing.assert_array_equal(hit_world_pos, world_pos)

  def test_raycast_hitting_empty_space(self):
    def mock_select(
        m, d, vopt, aspectratio, relx, rely, scn, selpnt, geomid, skinid):
      del (m, d, vopt, aspectratio, relx, rely, scn, selpnt, geomid,
           skinid)  # Unused.
      mock_body_id = -1  # Nothing selected.
      return mock_body_id
    with mock.patch(renderer.__name__ + '.mjlib') as mock_mjlib:
      mock_mjlib.mjv_select = mock.MagicMock(side_effect=mock_select)
      hit_body_id, hit_world_pos = self.camera.raycast(self.viewport, [0, 0])
      self.assertEqual(-1, hit_body_id)
      self.assertIsNone(hit_world_pos)

  def test_raycast_maps_coordinates_to_viewport_space(self):
    def build_mjv_select(expected_aspect_ratio, expected_viewport_pos):
      def mock_select(
          m, d, vopt, aspectratio, relx, rely, scn, selpnt, geomid, skinid):
        del m, d, vopt, scn, selpnt, geomid, skinid  # Unused.
        self.assertEqual(expected_aspect_ratio, aspectratio)
        np.testing.assert_array_equal(expected_viewport_pos, [relx, rely])
        mock_body_id = 0
        return mock_body_id
      return mock_select

    viewport_pos = [.5, .5]
    self.viewport.screen_to_inverse_viewport.return_value = viewport_pos
    mock_select = build_mjv_select(self.viewport.aspect_ratio, viewport_pos)

    with mock.patch(renderer.__name__ + '.mjlib') as mock_mjlib:
      mock_mjlib.mjv_select = mock.MagicMock(side_effect=mock_select)
      self.camera.raycast(self.viewport, [50, 25])

  def test_raycasts_disabled_when_camera_is_not_initialized(self):
    self.initialize_camera(False)
    hit_body_id, hit_world_pos = self.camera.raycast(self.viewport, [0, 0])
    self.assertEqual(-1, hit_body_id)
    self.assertIsNone(hit_world_pos)


class ViewportTest(parameterized.TestCase):

  def setUp(self):
    super(ViewportTest, self).setUp()
    self.viewport = renderer.Viewport()
    self.viewport.set_size(100, 100)

  @parameterized.parameters(
      ([0, 0], [0., 0.]),
      ([100, 0], [1., 0.]),
      ([0, 100], [0., 1.]),
      ([50, 50], [.5, .5]))
  def test_screen_to_viewport(self, screen_coords, viewport_coords):
    np.testing.assert_array_equal(
        viewport_coords, self.viewport.screen_to_viewport(screen_coords))

  @parameterized.parameters(
      ([0, 0], [0., 1.]),
      ([100, 0], [1., 1.]),
      ([0, 100], [0., 0.]),
      ([50, 50], [.5, .5]))
  def test_screen_to_inverse_viewport(self, screen_coords, viewport_coords):
    np.testing.assert_array_equal(
        viewport_coords,
        self.viewport.screen_to_inverse_viewport(screen_coords))

  @parameterized.parameters(
      ([10, 10], 1.),
      ([30, 40], 3./4.))
  def test_aspect_ratio(self, screen_size, aspect_ratio):
    self.viewport.set_size(screen_size[0], screen_size[1])
    self.assertEqual(aspect_ratio, self.viewport.aspect_ratio)


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Deterministic variations."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control.composer.variation import base


class Constant(base.Variation):
  """Wraps a constant value into a Variation object.

  This class is provided mainly for use in tests, to check that variations are
  invoked correctly without having to introduce randomness in test cases.
  """

  def __init__(self, value):
    self._value = value

  def __call__(self, initial_value=None, current_value=None, random_state=None):
    return self._value


class Sequence(base.Variation):
  """Variation representing a fixed sequence of values."""

  def __init__(self, values):
    self._values = values
    self._iterator = iter(self._values)

  def __call__(self, initial_value=None, current_value=None, random_state=None):
    try:
      return next(self._iterator)
    except StopIteration:
      self._iterator = iter(self._values)
      return next(self._iterator)



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Module containing the standard Jaco hand."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os

from dm_control import composer
from dm_control import mjcf
from dm_control.composer.observation import observable
from dm_control.entities.manipulators import base
from dm_control.entities.manipulators.kinova import assets_path

_JACO_HAND_XML_PATH = os.path.join(assets_path.KINOVA_ROOT, 'jaco_hand.xml')
_HAND_BODY = 'hand'
_PINCH_SITE = 'pinchsite'
_GRIP_SITE = 'gripsite'


class JacoHand(base.RobotHand):
  """A composer entity representing a Jaco hand."""

  def _build(self,
             name=None,
             use_pinch_site_as_tcp=False):
    """Initializes the JacoHand.

    Args:
      name: String, the name of this robot. Used as a prefix in the MJCF name
        name attributes.
      use_pinch_site_as_tcp: (optional) A boolean, if `True` the pinch site
        will be used as the tool center point. If `False` the grip site is used.
    """
    self._mjcf_root = mjcf.from_path(_JACO_HAND_XML_PATH)
    if name:
      self._mjcf_root.model = name
    # Find MJCF elements that will be exposed as attributes.
    self._bodies = self.mjcf_model.find_all('body')
    self._tool_center_point = self._mjcf_root.find(
        'site', _PINCH_SITE if use_pinch_site_as_tcp else _GRIP_SITE)
    self._joints = self._mjcf_root.find_all('joint')
    self._hand_geoms = list(self._mjcf_root.find('body', _HAND_BODY).geom)
    self._finger_geoms = [geom for geom in self._mjcf_root.find_all('geom')
                          if geom.name and geom.name.startswith('finger')]
    self._grip_site = self._mjcf_root.find('site', _GRIP_SITE)
    self._pinch_site = self._mjcf_root.find('site', _PINCH_SITE)

    # Add actuators.
    self._finger_actuators = [
        _add_velocity_actuator(joint) for joint in self._joints]

  def _build_observables(self):
    return JacoHandObservables(self)

  @property
  def tool_center_point(self):
    """Tool center point for the Jaco hand."""
    return self._tool_center_point

  @property
  def joints(self):
    """List of joint elements."""
    return self._joints

  @property
  def actuators(self):
    """List of finger actuators."""
    return self._finger_actuators

  @property
  def hand_geom(self):
    """List of geoms belonging to the hand."""
    return self._hand_geoms

  @property
  def finger_geoms(self):
    """List of geoms belonging to the fingers."""
    return self._finger_geoms

  @property
  def grip_site(self):
    """Grip site."""
    return self._grip_site

  @property
  def pinch_site(self):
    """Pinch site."""
    return self._pinch_site

  @property
  def pinch_site_pos_sensor(self):
    """Sensor that returns the cartesian position of the pinch site."""
    return self._pinch_site_pos_sensor

  @property
  def pinch_site_quat_sensor(self):
    """Sensor that returns the orientation of the pinch site as a quaternion."""
    return self._pinch_site_quat_sensor

  @property
  def mjcf_model(self):
    """Returns the `mjcf.RootElement` object corresponding to this robot."""
    return self._mjcf_root

  def set_grasp(self, physics, close_factors):
    """Sets the finger position to the desired positions.

    Args:
      physics: An instance of `mjcf.Physics`.
      close_factors: A number or list of numbers defining the desired grasp
        position of each finger. A value of 0 corresponds to fully opening a
        finger, while a value of 1 corresponds to fully closing it. If a single
        number is specified, the same position is applied to all fingers.
    """
    if not isinstance(close_factors, collections.Iterable):
      close_factors = (close_factors,) * len(self.joints)
    for joint, finger_factor in zip(self.joints, close_factors):
      joint_mj = physics.bind(joint)
      min_value, max_value = joint_mj.range
      joint_mj.qpos = min_value + (max_value - min_value) * finger_factor
    physics.after_reset()

    # Set target joint velocities to zero.
    physics.bind(self.actuators).ctrl = 0


def _add_velocity_actuator(joint):
  """Adds a velocity actuator to a joint, returns the new MJCF element."""
  # These parameters were adjusted to achieve a grip force of ~25 N and a finger
  # closing time of ~1.2 s, as specified in the datasheet for the hand.
  gain = 10.
  forcerange = (-1., 1.)
  ctrlrange = (-5., 5.)  # Based on Kinova's URDF.
  return joint.root.actuator.add(
      'velocity',
      joint=joint,
      name=joint.name,
      kv=gain,
      ctrllimited=True,
      ctrlrange=ctrlrange,
      forcelimited=True,
      forcerange=forcerange)


class JacoHandObservables(base.JointsObservables):
  """Observables for the Jaco hand."""

  @composer.observable
  def pinch_site_pos(self):
    """The position of the pinch site, in global coordinates."""
    return observable.MJCFFeature('xpos', self._entity.pinch_site)

  @composer.observable
  def pinch_site_rmat(self):
    """The rotation matrix of the pinch site in global coordinates."""
    return observable.MJCFFeature('xmat', self._entity.pinch_site)




# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for `dm_control.mjcf.physics`."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import os

# Internal dependencies.
from absl.testing import absltest
from absl.testing import parameterized
from dm_control import mjcf
from dm_control.mjcf import physics as mjcf_physics
from dm_control.mujoco.wrapper import mjbindings
import mock
import numpy as np
import six
from six.moves import cPickle
from six.moves import range
from six.moves import zip

mjlib = mjbindings.mjlib

ARM_MODEL = os.path.join(os.path.dirname(__file__), 'test_assets/robot_arm.xml')


class PhysicsTest(parameterized.TestCase):
  """Tests for `mjcf.Physics`."""

  def setUp(self):
    super(PhysicsTest, self).setUp()
    self.model = mjcf.from_path(ARM_MODEL)
    self.physics = mjcf.Physics.from_xml_string(
        self.model.to_xml_string(), assets=self.model.get_assets())
    self.random = np.random.RandomState(0)

  def sample_elements(self, namespace, single_element):
    all_elements = self.model.find_all(namespace)
    if single_element:
      # A single randomly chosen element from this namespace.
      elements = self.random.choice(all_elements)
      full_identifiers = elements.full_identifier
    else:
      # A random permutation of all elements in this namespace.
      elements = self.random.permutation(all_elements)
      full_identifiers = [element.full_identifier for element in elements]
    return elements, full_identifiers

  def test_construct_and_reload_from_mjcf_model(self):
    physics = mjcf.Physics.from_mjcf_model(self.model)
    physics.data.time = 1.
    physics.reload_from_mjcf_model(self.model)
    self.assertEqual(physics.data.time, 0.)

  @parameterized.parameters(
      # namespace, single_element
      ('geom', True),
      ('geom', False),
      ('joint', True),
      ('joint', False))
  def test_id(self, namespace, single_element):
    elements, full_identifiers = self.sample_elements(namespace, single_element)
    actual = self.physics.bind(elements).element_id
    if single_element:
      expected = self.physics.model.name2id(full_identifiers, namespace)
    else:
      expected = [self.physics.model.name2id(name, namespace)
                  for name in full_identifiers]
    np.testing.assert_array_equal(expected, actual)

  def assertCanGetAndSetBindingArray(
      self, binding, attribute_name, named_indexer, full_identifiers):

    # Read the values using the binding attribute.
    actual = getattr(binding, attribute_name)

    # Read them using the normal named indexing machinery.
    expected = named_indexer[full_identifiers]
    np.testing.assert_array_equal(expected, actual)

    # Assign an array of unique values to the attribute.
    expected = np.arange(actual.size, dtype=actual.dtype).reshape(actual.shape)
    setattr(binding, attribute_name, expected)

    # Read the values back using the normal named indexing machinery.
    actual = named_indexer[full_identifiers]
    np.testing.assert_array_equal(expected, actual)

  @parameterized.parameters(
      # namespace, attribute_name, model_or_data, field_name, single_element
      ('geom', 'xpos', 'data', 'geom_xpos', True),
      ('geom', 'xpos', 'data', 'geom_xpos', False),
      ('joint', 'qpos', 'data', 'qpos', True),
      ('joint', 'qpos', 'data', 'qpos', False),
      ('site', 'rgba', 'model', 'site_rgba', True),
      ('site', 'rgba', 'model', 'site_rgba', False),
      ('sensor', 'sensordata', 'data', 'sensordata', True),
      ('sensor', 'sensordata', 'data', 'sensordata', False))
  def test_attribute_access(self, namespace, attribute_name, model_or_data,
                            field_name, single_element):

    elements, full_identifiers = self.sample_elements(namespace, single_element)
    named_indexer = getattr(getattr(self.physics.named, model_or_data),
                            field_name)
    binding = self.physics.bind(elements)
    self.assertCanGetAndSetBindingArray(
        binding, attribute_name, named_indexer, full_identifiers)

  @parameterized.parameters(
      # namespace, attribute_name, model_or_data, field_name, single_element,
      # column_index
      ('geom', 'pos', 'model', 'geom_pos', True, None),
      ('geom', 'pos', 'model', 'geom_pos', False, None),
      ('geom', 'pos', 'model', 'geom_pos', True, 1),
      ('geom', 'pos', 'model', 'geom_pos', False, 1),
      ('geom', 'pos', 'model', 'geom_pos', True, 'y'),
      ('geom', 'pos', 'model', 'geom_pos', False, 'y'),
      ('geom', 'pos', 'model', 'geom_pos', True, slice(0, None, 2)),
      ('geom', 'pos', 'model', 'geom_pos', False, slice(0, None, 2)),
      ('geom', 'pos', 'model', 'geom_pos', True, [0, 2]),
      ('geom', 'pos', 'model', 'geom_pos', False, [0, 2]),
      ('geom', 'pos', 'model', 'geom_pos', True, ['x', 'z']),
      ('geom', 'pos', 'model', 'geom_pos', False, ['x', 'z']),
      ('joint', 'qpos', 'data', 'qpos', True, None),
      ('joint', 'qpos', 'data', 'qpos', False, None))
  def test_indexing(self, namespace, attribute_name, model_or_data,
                    field_name, single_element, column_index):

    elements, full_identifiers = self.sample_elements(namespace, single_element)
    named_indexer = getattr(getattr(self.physics.named, model_or_data),
                            field_name)
    binding = self.physics.bind(elements)

    if column_index is not None:
      binding_index = (attribute_name, column_index)
      try:
        named_index = np.ix_(full_identifiers, column_index)
      except ValueError:
        named_index = (full_identifiers, column_index)
    else:
      binding_index = attribute_name
      named_index = full_identifiers

    # Read the values by indexing the binding.
    actual = binding[binding_index]

    # Read them using the normal named indexing machinery.
    expected = named_indexer[named_index]
    np.testing.assert_array_equal(expected, actual)

    # Write an array of unique values into the binding.
    expected = np.arange(actual.size, dtype=actual.dtype).reshape(actual.shape)
    binding[binding_index] = expected

    # Read the values back using the normal named indexing machinery.
    actual = named_indexer[named_index]
    np.testing.assert_array_equal(expected, actual)

  def test_bind_mocap_body(self):
    pos = [1, 2, 3]
    quat = [1, 0, 0, 0]
    model = mjcf.RootElement()
    # Bodies are non-mocap by default.
    normal_body = model.worldbody.add('body', pos=pos, quat=quat)
    mocap_body = model.worldbody.add('body', pos=pos, quat=quat, mocap='true')
    physics = mjcf.Physics.from_xml_string(model.to_xml_string())

    binding = physics.bind(mocap_body)
    np.testing.assert_array_equal(pos, binding.mocap_pos)
    np.testing.assert_array_equal(quat, binding.mocap_quat)

    new_pos = [4, 5, 6]
    new_quat = [0, 1, 0, 0]
    binding.mocap_pos = new_pos
    binding.mocap_quat = new_quat
    np.testing.assert_array_equal(
        new_pos, physics.named.data.mocap_pos[mocap_body.full_identifier])
    np.testing.assert_array_equal(
        new_quat, physics.named.data.mocap_quat[mocap_body.full_identifier])

    with self.assertRaises(AttributeError):
      _ = physics.bind(normal_body).mocap_pos

    with six.assertRaisesRegex(
        self,
        ValueError,
        'Cannot bind to a collection containing multiple element types'):
      physics.bind([mocap_body, normal_body])

  def test_bind_worldbody(self):
    expected_mass = 10
    model = mjcf.RootElement()
    child = model.worldbody.add('body')
    child.add('geom', type='sphere', size=[0.1], mass=expected_mass)
    physics = mjcf.Physics.from_mjcf_model(model)
    mass = physics.bind(model.worldbody).subtreemass
    self.assertEqual(mass, expected_mass)

  def test_caching(self):
    all_joints = self.model.find_all('joint')

    original = self.physics.bind(all_joints)
    cached = self.physics.bind(all_joints)
    self.assertIs(cached, original)

    different_order = self.physics.bind(all_joints[::-1])
    self.assertIsNot(different_order, original)

    # Reloading the `Physics` instance should clear the cache.
    self.physics.reload_from_xml_string(
        self.model.to_xml_string(), assets=self.model.get_assets())
    after_reload = self.physics.bind(all_joints)
    self.assertIsNot(after_reload, original)

  def test_exceptions(self):
    joint = self.model.find_all('joint')[0]
    geom = self.model.find_all('geom')[0]
    with six.assertRaisesRegex(
        self,
        ValueError,
        'Cannot bind to a collection containing multiple element types'):
      self.physics.bind([joint, geom])

    with six.assertRaisesRegex(self, ValueError, 'cannot be bound to physics'):
      mjcf.physics.Binding(self.physics, 'invalid_namespace', 'whatever')

    binding = self.physics.bind(joint)
    with six.assertRaisesRegex(self, AttributeError, 'does not have attribute'):
      getattr(binding, 'invalid_attribute')

  def test_dirty(self):
    self.physics.forward()
    self.assertFalse(self.physics.is_dirty)

    joints, _ = self.sample_elements('joint', single_element=False)
    sites, _ = self.sample_elements('site', single_element=False)

    # Accessing qpos shouldn't trigger a recalculation.
    _ = self.physics.bind(joints).qpos
    self.assertFalse(self.physics.is_dirty)

    # Reassignments to qpos should cause the physics to become dirty.
    site_xpos_before = copy.deepcopy(self.physics.bind(sites).xpos)
    self.physics.bind(joints).qpos += 0.5
    self.assertTrue(self.physics.is_dirty)

    # Accessing stuff in mjModel shouldn't trigger a recalculation.
    _ = self.physics.bind(sites).pos
    self.assertTrue(self.physics.is_dirty)

    # Accessing stuff in mjData should trigger a recalculation.
    actual_sites_xpos_after = copy.deepcopy(self.physics.bind(sites).xpos)
    self.assertFalse(self.physics.is_dirty)
    self.assertFalse((actual_sites_xpos_after == site_xpos_before).all())

    # Automatic recalculation should render `forward` a no-op here.
    self.physics.forward()
    expected_sites_xpos_after = self.physics.bind(sites).xpos
    np.testing.assert_array_equal(actual_sites_xpos_after,
                                  expected_sites_xpos_after)

    # `forward` should not be called on subsequent queries to xpos.
    with mock.patch.object(
        self.physics, 'forward',
        side_effect=self.physics.forward) as mock_forward:
      _ = self.physics.bind(sites).xpos
      mock_forward.assert_not_called()

  @parameterized.parameters(True, False)
  def test_assign_while_dirty(self, assign_via_slice):
    actuators = self.model.find_all('actuator')
    if assign_via_slice:
      self.physics.bind(actuators).ctrl[:] = 0.75
    else:
      self.physics.bind(actuators).ctrl = 0.75
    self.assertTrue(self.physics.is_dirty)
    self.physics.step()
    self.assertTrue(self.physics.is_dirty)
    sensors = self.model.find_all('sensor')
    if assign_via_slice:
      self.physics.bind(sensors).sensordata[:] = 12345
    else:
      self.physics.bind(sensors).sensordata = 12345
    self.assertFalse(self.physics.is_dirty)
    np.testing.assert_array_equal(
        self.physics.bind(sensors).sensordata,
        [12345] * len(self.physics.bind(sensors).sensordata))

  def test_setitem_on_binding_attr(self):
    bodies, _ = self.sample_elements('body', single_element=False)
    xfrc_binding = self.physics.bind(bodies).xfrc_applied

    xfrc_binding[:, 1] = list(range(len(bodies)))
    for i, body in enumerate(bodies):
      self.assertEqual(xfrc_binding[i, 1], i)
      self.assertEqual(
          self.physics.named.data.xfrc_applied[body.full_identifier][1], i)

    xfrc_binding[:, 1] *= 2
    for i, body in enumerate(bodies):
      self.assertEqual(xfrc_binding[i, 1], 2 * i)
      self.assertEqual(
          self.physics.named.data.xfrc_applied[body.full_identifier][1], 2 * i)

    xfrc_binding[[1, 3, 5], 2] = 42
    for i, body in enumerate(bodies):
      actual_value = (
          self.physics.named.data.xfrc_applied[body.full_identifier][2])
      if i in [1, 3, 5]:
        self.assertEqual(actual_value, 42)
      else:
        self.assertNotEqual(actual_value, 42)

    # Bind to a single element.
    single_binding = self.physics.bind(bodies[0]).xfrc_applied
    single_binding[:2] = 55
    np.testing.assert_array_equal(single_binding[:2], [55, 55])
    np.testing.assert_array_equal(
        self.physics.named.data.xfrc_applied[bodies[0].full_identifier][:2],
        [55, 55])

  def test_empty_binding(self):
    binding = self.physics.bind([])
    self.assertEqual(binding.xpos.shape, (0,))
    with self.assertRaisesWithLiteralMatch(
        ValueError, 'Cannot assign a value to an empty binding.'):
      binding.xpos = 5

  @parameterized.parameters([('data', 'act'), ('data', 'act_dot')])
  def test_actuator_state_binding(self, model_or_data, attribute_name):

    def make_model_with_mixed_actuators():
      actuators = []
      is_stateful = []
      root = mjcf.RootElement()
      body = root.worldbody.add('body')
      body.add('geom', type='sphere', size=[0.1])
      slider = body.add('joint', type='slide', name='slide_joint')
      # Third-order `general` actuator.
      actuators.append(
          root.actuator.add(
              'general', dyntype='integrator', biastype='affine',
              dynprm=[1, 0, 0], joint=slider, name='general_act'))
      is_stateful.append(True)
      # Cylinder actuators are also third-order.
      actuators.append(
          root.actuator.add('cylinder', joint=slider, name='cylinder_act'))
      is_stateful.append(True)
      # A second-order actuator, added after the third-order actuators. The
      # actuators will be automatically reordered in the generated XML so that
      # the second-order actuator comes first.
      actuators.append(
          root.actuator.add('velocity', joint=slider, name='velocity_act'))
      is_stateful.append(False)
      return root, actuators, is_stateful

    model, actuators, is_stateful = make_model_with_mixed_actuators()
    physics = mjcf.Physics.from_mjcf_model(model)
    binding = physics.bind(actuators)
    named_indexer = getattr(
        getattr(physics.named, model_or_data), attribute_name)
    stateful_actuator_names = [
        actuator.full_identifier
        for actuator, stateful in zip(actuators, is_stateful) if stateful]
    self.assertCanGetAndSetBindingArray(
        binding, attribute_name, named_indexer, stateful_actuator_names)

  def test_bind_stateless_actuators_only(self):
    actuators = []
    root = mjcf.RootElement()
    body = root.worldbody.add('body')
    body.add('geom', type='sphere', size=[0.1])
    slider = body.add('joint', type='slide', name='slide_joint')
    actuators.append(
        root.actuator.add('velocity', joint=slider, name='velocity_act'))
    actuators.append(
        root.actuator.add('motor', joint=slider, name='motor_act'))
    # `act` should be an empty array if there are no stateful actuators.
    physics = mjcf.Physics.from_mjcf_model(root)
    self.assertEqual(physics.bind(actuators).act.shape, (0,))

  def make_simple_model(self):
    def add_submodel(root):
      body = root.worldbody.add('body')
      geom = body.add('geom', type='ellipsoid', size=[0.1, 0.2, 0.3])
      site = body.add('site', type='sphere', size=[0.1])
      return body, geom, site
    root = mjcf.RootElement()
    add_submodel(root)
    add_submodel(root)
    return root

  def quat2mat(self, quat):
    result = np.empty(9, dtype=np.double)
    mjlib.mju_quat2Mat(result, np.asarray(quat))
    return result

  @parameterized.parameters(['body', 'geom', 'site'])
  def test_write_to_pos(self, entity_type):
    root = self.make_simple_model()
    entity1, entity2 = root.find_all(entity_type)
    physics = mjcf.Physics.from_mjcf_model(root)
    first = physics.bind(entity1)
    second = physics.bind(entity2)

    # Initially both entities should be 'sameframe'
    self.assertEqual(first.sameframe, 1)
    self.assertEqual(second.sameframe, 1)

    # Assigning to `pos` should disable 'sameframe' only for that entity.
    new_pos = (0., 0., 0.1)
    first.pos = new_pos
    self.assertEqual(first.sameframe, 0)
    self.assertEqual(second.sameframe, 1)
    # `xpos` should reflect the new position.
    np.testing.assert_array_equal(first.xpos, new_pos)

    # Writing into the `pos` array should also disable 'sameframe'.
    new_x = -0.1
    pos_array = second.pos
    pos_array[0] = new_x
    self.assertEqual(second.sameframe, 0)
    # `xpos` should reflect the new position.
    self.assertEqual(second.xpos[0], new_x)

  @parameterized.parameters(['body', 'geom', 'site'])
  def test_write_to_quat(self, entity_type):
    root = self.make_simple_model()
    entity1, entity2 = root.find_all(entity_type)
    physics = mjcf.Physics.from_mjcf_model(root)
    first = physics.bind(entity1)
    second = physics.bind(entity2)

    # Initially both entities should be 'sameframe'
    self.assertEqual(first.sameframe, 1)
    self.assertEqual(second.sameframe, 1)

    # Assigning to `quat` should disable 'sameframe' only for that entity.
    new_quat = (0., 0., 0., 1.)
    first.quat = new_quat
    self.assertEqual(first.sameframe, 0)
    self.assertEqual(second.sameframe, 1)
    # `xmat` should reflect the new quaternion.
    np.testing.assert_allclose(first.xmat, self.quat2mat(new_quat))

    # Writing into the `quat` array should also disable 'sameframe'.
    new_w = -1.
    quat_array = second.quat
    quat_array[0] = new_w
    self.assertEqual(second.sameframe, 0)
    # `xmat` should reflect the new quaternion.
    np.testing.assert_allclose(second.xmat, self.quat2mat(quat_array))

  def test_write_to_ipos(self):
    root = self.make_simple_model()
    entity1, entity2 = root.find_all('body')
    physics = mjcf.Physics.from_mjcf_model(root)
    first = physics.bind(entity1)
    second = physics.bind(entity2)

    # Initially both bodies should be 'simple' and 'sameframe'
    self.assertEqual(first.simple, 1)
    self.assertEqual(first.sameframe, 1)
    self.assertEqual(second.simple, 1)
    self.assertEqual(second.sameframe, 1)

    # Assigning to `ipos` should disable 'simple' and 'sameframe' only for that
    # body.
    new_ipos = (0., 0., 0.1)
    first.ipos = new_ipos
    self.assertEqual(first.simple, 0)
    self.assertEqual(first.sameframe, 0)
    self.assertEqual(second.simple, 1)
    self.assertEqual(second.sameframe, 1)
    # `xipos` should reflect the new position.
    np.testing.assert_array_equal(first.xipos, new_ipos)

    # Writing into the `ipos` array should also disable 'simple' and
    # 'sameframe'.
    new_x = -0.1
    ipos_array = second.ipos
    ipos_array[0] = new_x
    self.assertEqual(second.simple, 0)
    self.assertEqual(second.sameframe, 0)
    # `xipos` should reflect the new position.
    self.assertEqual(second.xipos[0], new_x)

  def test_write_to_iquat(self):
    root = self.make_simple_model()
    entity1, entity2 = root.find_all('body')
    physics = mjcf.Physics.from_mjcf_model(root)
    first = physics.bind(entity1)
    second = physics.bind(entity2)

    # Initially both bodies should be 'simple' and 'sameframe'
    self.assertEqual(first.simple, 1)
    self.assertEqual(first.sameframe, 1)
    self.assertEqual(second.simple, 1)
    self.assertEqual(second.sameframe, 1)

    # Assigning to `iquat` should disable 'simple' and 'sameframe' only for that
    # body.
    new_iquat = (0., 0., 0., 1.)
    first.iquat = new_iquat
    self.assertEqual(first.simple, 0)
    self.assertEqual(first.sameframe, 0)
    self.assertEqual(second.simple, 1)
    self.assertEqual(second.sameframe, 1)
    # `ximat` should reflect the new quaternion.
    np.testing.assert_allclose(first.ximat, self.quat2mat(new_iquat))

    # Writing into the `iquat` array should also disable 'simple' and
    # 'sameframe'.
    new_w = -0.1
    iquat_array = second.iquat
    iquat_array[0] = new_w
    self.assertEqual(second.simple, 0)
    self.assertEqual(second.sameframe, 0)
    # `ximat` should reflect the new quaternion.
    np.testing.assert_allclose(second.ximat, self.quat2mat(iquat_array))

  @parameterized.parameters([dict(order='C'), dict(order='F')])
  def test_copy_synchronizing_array_wrapper(self, order):
    root = self.make_simple_model()
    physics = mjcf.Physics.from_mjcf_model(root)
    xpos_view = physics.bind(root.find_all('body')).xpos
    xpos_view_copy = xpos_view.copy(order=order)

    np.testing.assert_array_equal(xpos_view, xpos_view_copy)
    self.assertFalse(np.may_share_memory(xpos_view, xpos_view_copy),
                     msg='Original and copy should not share memory.')
    self.assertIs(type(xpos_view_copy), np.ndarray)

    # Check that `order=` is respected.
    if order == 'C':
      self.assertTrue(xpos_view_copy.flags.c_contiguous)
      self.assertFalse(xpos_view_copy.flags.f_contiguous)
    elif order == 'F':
      self.assertFalse(xpos_view_copy.flags.c_contiguous)
      self.assertTrue(xpos_view_copy.flags.f_contiguous)

    # The copy should be writeable.
    self.assertTrue(xpos_view_copy.flags.writeable)
    new_value = 99.
    xpos_view_copy[0, -1] = new_value
    self.assertEqual(xpos_view_copy[0, -1], new_value)

  def test_error_when_pickling_synchronizing_array_wrapper(self):
    root = self.make_simple_model()
    physics = mjcf.Physics.from_mjcf_model(root)
    xpos_view = physics.bind(root.find_all('body')).xpos
    with self.assertRaisesWithLiteralMatch(
        NotImplementedError,
        mjcf_physics._PICKLING_NOT_SUPPORTED.format(type=type(xpos_view))):
      cPickle.dumps(xpos_view)

if __name__ == '__main__':
  absltest.main()
